/* $Header: /nw/tony/src/stevie/src/RCS/cmdline.c,v 1.20 89/08/13 11:41:23 tony Exp $
 *
 * Routines to parse and execute "command line" commands, such as searches
 * or colon commands.
 */

#include "stevie.h"

static	uchar	*altfile = NULL;	/* alternate file */
static	int	altline;		/* line # in alternate file */

static	uchar	*nowrtmsg = (uchar*) 
			"No write since last change (use ! to override)";
static	uchar	*nooutfile = (uchar*) "No output file";
static	uchar	*morefiles = (uchar*) "%d more file%s to edit";
static	uchar	*nomore =    (uchar*) "No more files to edit";
static	uchar	*toomany =   (uchar*) "Too many file names";

static	uchar	*badmsg;

static	bool_t	standout = FALSE;

extern	char	**files;		/* used for "n" and "rew" */
extern	int	numfiles, curfile;
void doinitfile();

static	bool_t	doecmd();
static	void	get_range();
static	LPTR	*get_line();

/* Position entry for tag stack */
typedef struct Tag_entry {
	char 	*file;			/* file name */
	long	line;			/* user's line number */
	struct Tag_entry *next;
} Tag_entry;

/* Tag stack-related operations */
static Tag_entry	*tag_new();
static void		tag_dispose();
static void  		tag_pop();
static bool_t  		tag_push();
static void		tag_list();

static Tag_entry  *Tag_stack = NULL;

uchar *
pretty(str, colp)
register uchar	*str;
register int	*colp;
{
	register uchar	*p, *q;
	static uchar	buf[MAXMLEN];

	p = buf;
	while (*str) {
#ifdef KANJI
		if (isKanji(*str)) {
			*p++ = *str++;
			*p++ = *str++;
			*colp += 2;
			continue;
		}
#endif
		if (*str == TAB) {
			do {
				*p++ = ' ';
			} while (++(*colp) % P(P_TS));
			str++;
		} else if (chars[*str].ch_size > 1) {
			q = (uchar *)chars[*str].ch_str;
			while (*q) {
				*p++ = *q++;
				(*colp)++;
			}
			str++;
		} else {
			*p++ = *str++;
			(*colp)++;
		}
	}
	*p = NUL;
	return buf;
}

/*
 * getcmdln() - read a command line from the terminal
 *
 * Reads a command line started by typing '/', '?', '!', or ':'. Returns a
 * pointer to the string that was read. For searches, an optional trailing
 * '/' or '?' is removed.
 */
uchar *
getcmdln(firstc)
uchar	firstc;
{
	static	uchar	buff[MAXCLIN];
	register uchar	*p = buff;
	register int	c;
	register uchar	*q;
	int	col = 1;

	State = CMDLINE;
	gotocmd(TRUE, firstc);

	can_ab = TRUE;

	/* collect the command string, handling '\b' and @ */
	for (;;) {
		c = vgetc();
		switch (c) {

		default:		/* a normal character */
			if (c == DEL || 0x00 <= c && c < 0x20 &&
					 c != TAB && c != Ctrl('V')) {
				beep();
				break;
			}
			if (c == Ctrl('V')) {
				can_ab = FALSE;
				if ((c = vgetc()) == NL)
					goto newline;
			}
		normchar:
			if (p - buff > MAXCLIN - 2) {
				beep();
				break;
			}
			outstr(pretty(mkstr(c), &col));
#ifdef KANJI
			if (c & 0xff00)
				*p++ = (c >> 8) & 0xff;
			*p++ = c & 0xff;
			can_ab = ((c & 0xff00) || !isalnum(c) && c != '_');
#else
			*p++ = c;
			can_ab = (!isalnum(c) && c != '_');
#endif
			break;

		case BS:
		case DEL:
			if (p == buff) {
				State = NORMAL;
#ifdef FEPCTRL
				if (P(P_FC))
					fep_force_off();
#endif
				msg("");
				return NULL;		/* back to cmd mode */
			}
			/*
			 * this is gross, but it relies
			 * only on 'gotocmd'
			 */
			p--;
			gotocmd(TRUE, firstc);
			col = 1;
#ifdef KANJI
			for (q = buff; q < p; q++) {
				if (isKanji(*q)) {
					if (q + 1 == p) {
						p--;
						break;
					}
				}
			}
#endif
			*p = NUL;
			outstr(pretty(buff, &col));
			can_ab = TRUE;
			break;

		case Ctrl('U'):			/* line kill */
			p = buff;
			gotocmd(TRUE, firstc);
			col = 1;
			can_ab = TRUE;
			break;

		case NL:			/* done reading the line */
		case CR:
		newline:
			goto out;
		}
	}
out:
	*p = '\0';

	if (firstc == '/' || firstc == '?') {	/* did we do a search? */
		/*
		 * Look for a terminating '/' or '?'. This will be the first
		 * one that isn't quoted. Truncate the search string there.
		 */
		for (p = buff; *p ;) {
#ifdef KANJI
			if (isKanji(*p))	/* next char is kanji */
				p += 2;
			else
#endif
			if (*p == firstc) {	/* we're done */
				*p = '\0';
				break;
			} else if (*p == '\\')	/* next char quoted */
				p += 2;
			else
				p++;		/* normal char */
		}
	}
	State = NORMAL;
#ifdef FEPCTRL
	if (P(P_FC))
		fep_force_off();
#endif
	windgoto(Rows-1, 0);
	flushbuf();
	return buff;
}

/*
 * fexpand() - expand '%' in the argument into Filename, '#' into altfile
 */
static int
fexpand(arg, maxlen)
uchar	*arg;
int	maxlen;
{
	register uchar	*p, *q;
	register uchar	*start, *bp;
	uchar	*buf, *nam;
	bool_t	has_wildcard, scanning_word;    
	int	words;
	extern int glob();

	if (arg == NULL)
		return 0;
	if ((buf = alloc(maxlen + 1)) == NULL)
		return 0;
	if ((nam = alloc(maxlen + 1)) == NULL) {
		free(buf);
		return 0;
	}
	while ((p = (uchar *)strchr(arg, '%')) && (p==arg || p[-1]!='\\')) {
		if (Filename == NULL) {
			emsg("No filename");
			free(buf);
			free(nam);
			return 0;
		}
		*p = NUL;
		strcpy(buf, arg);
		strcat(buf, Filename);
		strcat(buf, p+1);
		strcpy(arg, buf);
	}
	while ((p = (uchar *)strchr(arg, '#')) && (p==arg || p[-1]!='\\')) {
		if (altfile == NULL) {
			emsg("No alternate file");
			free(buf);
			free(nam);
			return 0;
		}
		*p = NUL;
		strcpy(buf, arg);
		strcat(buf, altfile);
		strcat(buf, p+1);
		strcpy(arg, buf);
	}

	has_wildcard = FALSE;
	scanning_word = FALSE;
	p = arg;
	bp = buf;
	start = buf;
	do {
		switch (*p) {
		case '$':
		case '[':
		case '{':
		case '?':
		case '~':
		case ']':
		case '}':
		case '*':
			if (! scanning_word) {	/* start scanning a new word */
				scanning_word = TRUE;       
				start = bp;
			}
			has_wildcard = TRUE;
			*bp++ = *p;
			break;

		case ' ':
		case TAB:	/* whitespace delimits a "word" */
		case NUL:
			if (scanning_word) {
				/* glob only if it has wildcards */
				if (has_wildcard) {
					int i, n, len;

					*bp = NUL;
					len = maxlen - (start - buf) - 1;
					n = glob(start, nam, len);
					for (q=nam, i=0; i < n;  i++) {  
						/* replace NULs with spaces */
						q = (uchar *) strchr(q, NUL);  
						*q = ' ';
					}
					*q = NUL;
					strncpy(start, nam, len);
					bp = (uchar *) strchr(start, NUL);
					has_wildcard = FALSE;
				} else {
					*bp++ = *p;
				}
				scanning_word = FALSE;	/* end of word */
			} else {
				*bp++ = *p;	
			}
			break;
			
		default:
			if (! scanning_word) {	/* start scanning a new word */
				scanning_word = TRUE;       
				has_wildcard  = FALSE;
				start = bp;
			}
			*bp++ = *p;
			break;
		}
	} while (*p++ != NUL && ((bp-buf) < maxlen));
	buf[maxlen] = NUL;
	strncpy(arg, buf, maxlen);
	free(buf);
	free(nam);

	words = 0;
	p = arg;
	while (*p == ' ' || *p == TAB)
		p++;
	while (*p) {
		while (*p && *p != ' ' && *p != TAB) {
			if (*p == '\\' && p[1])
				p++;
			p++;
		}
		while (*p == ' ' || *p == TAB)
			p++;
		words++;
	}
	return words;
}

/*
 * docmdln() - handle a colon command
 *
 * Handles a colon command received interactively by getcmdln() or from
 * the environment variable "EXINIT" (or eventually .virc).
 */
void
docmdln(cmdline)
uchar	*cmdline;
{
	uchar	buff[MAXCLIN];
	uchar	cmdbuf[MAXCLIN];
	uchar	argbuf[MAXCLIN];
	uchar	*cmd, *arg;
	register uchar	*p;
	/*
	 * The next two variables contain the bounds of any range given in a
	 * command. If no range was given, both contain null line pointers.
	 * If only a single line was given, u_pos will contain a null line
	 * pointer.
	 */
	LPTR	l_pos, u_pos;

	/*
	 * Clear the range variables.
	 */
	l_pos.linep = (struct line *) NULL;
	u_pos.linep = (struct line *) NULL;

	if (cmdline == NULL)
		return;

	if (strlen(cmdline) > MAXCLIN-2) {
		emsg("Error: command line too long");
		return;
	}
	strcpy(buff, cmdline);

	/* skip any initial white space */
	for (cmd = buff; *cmd != NUL && (*cmd == ' ' || *cmd == TAB) ;cmd++)
		;

	if (*cmd == '%') {		/* change '%' to "1,$" */
		strcpy(cmdbuf, "1,$");	/* kind of gross... */
		strcat(cmdbuf, cmd+1);
		strcpy(cmd, cmdbuf);
	}

	/*
	 * Parse a range, if present (and update the cmd pointer).
	 */
	get_range(&cmd, &l_pos, &u_pos);

	if (badmsg) {
		/*
		 * range error found in lower routines
		 */
		emsg(badmsg);
		return;
	}

	if (l_pos.linep != NULL) {
		if (LINEOF(&l_pos) > LINEOF(&u_pos)) {
			emsg("Invalid range");
			return;
		}
	}

	strcpy(cmdbuf, cmd);	/* save the unmodified command */

	/* isolate the command and find any argument */
	for ( p=cmd; *p != NUL && *p != ' ' && *p != TAB; p++ )
		;
	if ( *p == NUL )
		arg = NULL;
	else {
		*p = NUL;
		for (p++; *p != NUL && (*p == ' ' || *p == TAB) ;p++)
			;
		if (*p == NUL)
			arg = NULL;
		else {
			strcpy(argbuf, p);
			arg = argbuf;
		}
	}
	if (*cmd == '"') {
		/* comment - do nothing */
		return;
	}
	if (strcmp(cmd,"q!") == 0)
		getout(TRUE);
	if (strcmp(cmd,"q") == 0) {
		if (Changed) {
			emsg(nowrtmsg);
			return;
		}
		if ((curfile + 1) < numfiles) {
			int	i = numfiles - curfile - 1;
			sprintf(buff, morefiles, i, (i > 1)? "s": "");
			emsg(buff);
			return;
		}
		getout(TRUE);
	}
	if (strcmp(cmd,"w") == 0 || strcmp(cmd,"w!") == 0) {
		if (arg == NULL) {
			if (Filename != NULL)
				writeit(Filename, &l_pos, &u_pos, cmd[1]=='!');
			else
				emsg(nooutfile);
			return;
		}
		if (fexpand(arg, MAXCLIN-1-(arg-argbuf)) == 1)
			writeit(arg, &l_pos, &u_pos, cmd[1]=='!');
		else
			emsg(toomany);
		return;
	}
	if (strcmp(cmd,"wq") == 0 || strcmp(cmd,"wq!") == 0) {
		if (Filename == NULL) {
			emsg(nooutfile);
			return;
		}
		if (writeit(Filename, (LPTR *)NULL, (LPTR *)NULL, cmd[2]=='!'))
			getout(TRUE);
		return;
	}
	if (strcmp(cmd, "x") == 0) {
		doxit(TRUE);
		return;
	}
	if (*cmd == 'n') {
		if ((P(P_AW))&&(!autowriteit(cmd[1] == '!')))
			return;
		if (arg == NULL) {
			if (curfile + 1 >= numfiles) {
				emsg(nomore);
				return;
			}
			curfile++;
		} else {
			char	**newfiles;
			uchar	*top;
			int	n, num;

			if ((p = alloc(GBSIZE)) == NULL)
				return;
			arg = (uchar *)strcpy(p, arg);
			if ((num = fexpand(arg, GBSIZE-1)) == 0) {
				free(arg);
				return;
			}
			newfiles = (char **)alloc(num * sizeof(char *));
			if (newfiles == NULL) {
				free(arg);
				return;
			}
			n = 0;
			p = arg;
			while (*p == TAB || *p == ' ')
				p++;
			while (*p) {
				top = p;
				while (*p && *p != TAB && *p != ' ') {
					if (*p == '\\' && p[1])
						p++;
					p++;
				}
				if (*p) {
					*p++ = NUL;
					while (*p == TAB || *p == ' ')
						p++;
				}
				newfiles[n] = (char *)strsave(top);
				if (newfiles[n] == NULL) {
					while (--n >= 0)
						free(newfiles[n]);
					free((char *)newfiles);
					free(arg);
					return;
				}
				n++;
			}
			if (num > 1) {
				sprintf(arg, "%d files to edit", num);
				msg(arg);
			}
			free(arg);
			arg = NULL;
			if (files) {
				for (n = 0; n < numfiles; n++)
					free(files[n]);
				free((char *)files);
			}
			files = newfiles;
			numfiles = num;
			curfile = 0;
		}
		/*
		 * stuff ":e[!] FILE\n"
		 */
		stuffin(":e");
		if (cmd[1] == '!')
			stuffin("!");
		stuffin(" ");
		if (arg != NULL)
			stuffin(arg);
		else
			stuffin(files[curfile]);
		stuffin("\n");
		return;
	}
	if (*cmd == 'N') {
		if (curfile > 0) {
			if ((P(P_AW))&&(!autowriteit(cmd[1] == '!')))
				return;
			/*
			 * stuff ":e[!] FILE\n"
			 */
			stuffin(":e");
			if (cmd[1] == '!')
				stuffin("!");
			stuffin(" ");
			stuffin(files[--curfile]);
			stuffin("\n");
		} else
			emsg(nomore);
		return;
	}
	if (strncmp(cmd, "rew", 3) == 0) {
		if (numfiles <= 1)		/* nothing to rewind */
			return;
		if ((P(P_AW))&&(!autowriteit(cmd[3] == '!')))
			return;
		curfile = 0;
		/*
		 * stuff ":e[!] FILE\n"
		 */
		stuffin(":e");
		if (cmd[3] == '!')
			stuffin("!");
		stuffin(" ");
		stuffin(files[0]);
		stuffin("\n");
		return;
	}
	if (strcmp(cmd,"e") == 0 || strcmp(cmd,"e!") == 0) {
		if (arg && fexpand(arg, MAXCLIN-1-(arg-argbuf)) == 0)
			return;
		(void) doecmd(arg, cmd[1] == '!');
		return;
	}
	/*
	 * The command ":e#" gets expanded to something like ":efile", so
	 * detect that case here.
	 */
	if (*cmd == 'e' && arg == NULL) {
		if (fexpand(cmd, MAXCLIN-1-(cmd-cmdbuf)) == 0)
			return;
		if (cmd[1] == '!')
			(void) doecmd(&cmd[2], TRUE);
		else
			(void) doecmd(&cmd[1], FALSE);
		return;
	}
	if (strcmp(cmd,"f") == 0) {
		if (arg == NULL) {
			fileinfo();
			return;
		}
		if (fexpand(arg, MAXCLIN-1-(arg-argbuf)) != 1) {
			emsg(toomany);
			return;
		}
		Filename = strsave(arg);
		filemess("");
		return;
	}
	if (cmd[0] == 'r' && cmd[1] == '!') {
		cmdread(cmdbuf+2);
		return;
	}
	if (strcmp(cmd,"r") == 0) {
		if (arg == NULL) {
			emsg("No filename to read");
			return;
		}
		if (arg[0] == '!') {
			cmdread(arg+1);
			return;
		}
		if (fexpand(arg, MAXCLIN-1-(arg-argbuf)) != 1) {
			emsg(toomany);
			return;
		}
		if (readfile(arg, Curschar, 1)) {
			emsg("No such file or directory");
			return;
		}
		updatescreen();
		CHANGED;
		return;
	}
	if (strcmp(cmd,"=") == 0) {
		if (l_pos.linep == NULL)
			l_pos.linep = Fileend->linep->prev;
		smsg("%d", cntllines(Filemem, &l_pos));
		return;
	}
	if (strncmp(cmd,"ta", 2) == 0) {
		if ((P(P_AW))&&(!autowriteit(cmd[2] == '!')))
			return;
		dotag(arg, cmd[2] == '!');
		return;
	}
	if (strncmp(cmd, "pop", 3) == 0) {
		if ((P(P_AW))&&(!autowriteit(cmd[3] == '!')))
			return;
		/*
		 * Pop the tag stack
		 */
		tag_pop(cmd[3] == '!');
		return;
	}
	if (strncmp(cmd,"set", 2) == 0) {
		doset(arg);
		return;
	}
	if (strcmp(cmd,"help") == 0) {
		if (help()) {
			screenclear();
			updatescreen();
		}
		return;
	}
	if (strncmp(cmd, "ve", 2) == 0) {
		extern	uchar	*Version;

		msg(Version);
		return;
	}
	if (strcmp(cmd, "sh") == 0) {
		doshell(NULL);
		return;
	}
	if (strncmp(cmd, "so", 2) == 0) {   /* source a file of commands */
		doinitfile(arg);	/* MFWyle 19mar90 */
		return;
	}
	if (*cmd == '!') {
		arg = cmdbuf + 1;
		if (fexpand(arg, MAXCLIN-1-(arg-cmdbuf)) == 0)
			return;
		doshell(arg);
		return;
	}
	if (*cmd == 's') {
		dosub(&l_pos, &u_pos, cmdbuf+1, NULL);
		return;
	}
	if (*cmd == 'g' || *cmd == 'v') {
		doglob(&l_pos, &u_pos, cmdbuf+1, (*cmd == 'v'));
		return;
	}
	if (strcmp(cmd, "ab") == 0 || strncmp(cmd, "una", 3) == 0) {
		domap(arg, ABBR, *cmd == 'a');
		return;
	}
	if (strncmp(cmd, "map", 3) == 0 || strncmp(cmd, "unm", 3) == 0) {
		int set, bang;
		set = (*cmd == 'm');
		bang = (set? cmd[3] == '!': cmd[5] == '!');
		domap(arg, bang? MAPI: MAPC, set);
		return;
	}
	/*
	 * simple minded ":m,nd", ":m,ny", ":m,n<" and ":m,n>" implementation
	 */
	if (*cmd == 'd' || *cmd == 'y' || *cmd == '<' || *cmd == '>') {
		sprintf(buff, "%dG%c%dG", cntllines(Filemem, &l_pos), *cmd,
					  cntllines(Filemem, &u_pos));
		stuffin(buff);
		return;
	}
	/*
	 * simple minded ":m,n{m,mo,t,co}x" (range move/copy) implementation
	 * originally by Eric Bowles.
	 * Simulate with:  `mG {d,y} nG xG p'
	 */
	if (*cmd == 'm' || *cmd == 't' || strncmp(cmd, "co", 2) == 0) {
		LPTR	dest_pos, dummy_pos;
		int	c, o, m, n, x;

		c = *cmd++;
		if (*cmd == 'o')
			cmd++;
		if (*cmd == NUL)
			cmd = arg;
		/* cmd now points to destination specifier */
		o = (c == 'm')? 'd': 'y';
		m = cntllines(Filemem, &l_pos);
		n = cntllines(Filemem, &u_pos);
		if (!strcmp(cmd, "0")) {
			x = 0;
		} else {
			dest_pos.linep = (struct line *) NULL;
			get_range(&cmd, &dest_pos, &dummy_pos);
			if (badmsg) {
				emsg(badmsg);
				return;
			}
			if (*cmd != NUL) {
				sprintf(buff,
					"Extra character at end of %s command",
					c=='m'? "move": c=='t'? "t": "copy");
				emsg(buff);
				return;
			}
			if (dest_pos.linep == NULL) {
				sprintf(buff, "%s requires a trailing address",
					c=='m'? "move": c=='t'? "t": "copy");
				emsg(buff);
				return;
			}
			x = cntllines(Filemem, &dest_pos);
		}
		if (c == 'm') {
			if (x == m - 1) {
				emsg("That move would do nothing!");
				return;
			} else if (m <= x && x <= n) {
				emsg("Move to a moved line");
				return;
			} else if (x > n) {
				x -= n - m + 1;
			}
		}
		if (x == 0)
			sprintf(buff, "%dG%c%dG1GP", m, o, n);
		else
			sprintf(buff, "%dG%c%dG%dGp", m, o, n, x);
		stuffin(buff);
		return;
	}

	/*
	 * If we got a line, but no command, then go to the line.
	 */
	if (*cmd == NUL && l_pos.linep != NULL) {
		*Curschar = l_pos;
		return;
	}

	emsg("Unrecognized command");
}

autowriteit(force)
bool_t	force;
{
	if (!Changed)
		return 1;
	if (Filename != NULL) {
		if (!writeit(Filename, (LPTR *)NULL, (LPTR *)NULL, FALSE))
			return force;
	} else {
		emsg(nooutfile);
		return 0;
	}
	return 1;
}

void
doinitfile(file)
char *file;
{
	FILE	*fp;
	uchar	*p, *q, buf[MAXCLIN];

	if ((fp = fopenb(file, "r")) == NULL)
		return;

	while (fgets(buf, MAXCLIN, fp) != NULL) {
#ifdef CPMSTYLE
		if (!strcmp(buf, mkstr(Ctrl('Z'))))
			continue;
#endif
		p = (uchar *)strchr(buf, NL);
		if (p == NULL)
			buf[MAXCLIN-1] = NUL;
#ifdef CPMSTYLE
		else if (p[-1] == CR)
			p[-1] = NUL;
#endif
		else
			*p = NUL;
		q = buf;
		for (p = buf; *p; p++) {
			if (*p == Ctrl('V') && p[1] != NUL)
				p++;
			*q++ = *p;
		}
		*q = NUL;
		docmdln(buf);
	}
	fclose(fp);
}

doxit(havemsg)
bool_t	havemsg;
{
	uchar	buff[MAXMLEN];
	int	i;

	if (!havemsg && Changed)
		havemsg = TRUE;
	if (!autowriteit(FALSE))
		return;
	if ((curfile + 1) < numfiles) {
		i = numfiles - curfile - 1;
		sprintf(buff, morefiles, i, (i > 1)? "s": "");
		emsg(buff);
	} else
		getout(havemsg);
}

/*
 * get_range - parse a range specifier
 *
 * Ranges are of the form:
 *
 * addr[,addr]
 *
 * where 'addr' is:
 *
 * $  [+- NUM]
 * 'x [+- NUM]	(where x denotes a currently defined mark)
 * .  [+- NUM]
 * NUM
 *
 * The pointer *cp is updated to point to the first character following
 * the range spec. If an initial address is found, but no second, the
 * upper bound is equal to the lower.
 */
static void
get_range(cp, lower, upper)
register uchar	**cp;
LPTR	*lower, *upper;
{
	register LPTR	*l;
	register uchar	*p;

	badmsg = NULL;

	if ((l = get_line(cp)) == NULL)
		return;

	*lower = *l;

	for (p = *cp; *p != NUL && isspace(*p) ;p++)
		;

	*cp = p;

	if (*p != ',') {		/* is there another line spec ? */
		*upper = *lower;
		return;
	}

	*cp = ++p;

	if ((l = get_line(cp)) == NULL) {
		*upper = *lower;
		return;
	}

	*upper = *l;
	return;
}

static LPTR *
get_line(cp)
uchar	**cp;
{
	static	LPTR	pos;
	LPTR	*lp;
	LPTR	*secretsearch();
	register uchar	*p, *q, c;
	register int	lnum;

	pos.index = 0;		/* shouldn't matter... check back later */

	p = *cp;
	/*
	 * Determine the basic form, if present.
	 */
	switch (c = *p++) {

	case '$':
		pos.linep = Fileend->linep->prev;
		break;

	case '.':
		pos.linep = Curschar->linep;
		break;

	case '\'':
		if ((lp = getmark(*p++)) == NULL) {
			badmsg = (uchar *)"Unknown mark";
			return (LPTR *) NULL;
		}
		pos = *lp;
		break;

	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
		for (lnum = c - '0'; isdigit(*p) ;p++)
			lnum = (lnum * 10) + (*p - '0');

		if ((lp = gotoline(lnum)) == NULL) {
			badmsg = (uchar *)"Not that many lines in buffer";
			return (LPTR *) NULL;
		}
		pos = *lp;
		break;

	case '/':	/* forward/backward search */
	case '?':
		q = p;
		for (; *p != c && *p != '\0'; p++) {
			if (*p == '\\')
				p++;
		}
		if (*p == c)
			*(p++) = '\0';
		if ((lp = secretsearch(c=='/'?FORWARD:BACKWARD, q)) == NULL) {
			return (LPTR*)NULL;
		}
		pos = *lp;
		break;		

	default:
		return (LPTR *) NULL;
	}

	while (*p != NUL && isspace(*p))
		p++;

	if (*p == '-' || *p == '+') {
		bool_t	neg = (*p++ == '-');

		for (lnum = 0; isdigit(*p) ;p++)
			lnum = (lnum * 10) + (*p - '0');

		if (neg)
			lnum = -lnum;

		if ((lnum = cntllines(Filemem, &pos) + lnum) < 1) {
			badmsg = (uchar *)
				"Negative address - first buffer line is 1";
			return (LPTR *) NULL;
		}
		if ((lp = gotoline(lnum)) == NULL) {
			badmsg = (uchar *)"Not that many lines in buffer";
			return (LPTR *) NULL;
		}
		pos = *lp;
	}

	*cp = p;
	return &pos;
}

/*
 * dotag(tag, force) - goto tag
 */
void
dotag(tag, force)
uchar	*tag;
bool_t	force;
{
	FILE	*tp, *fopen();
	uchar	lbuf[MAXTLIN];		/* line buffer */
	uchar	pbuf[MAXTLIN];		/* search pattern buffer */
	register uchar	*fname, *str;
	register uchar	*p;
	Tag_entry *cur_pos;

	if (tag == NULL) {
		tag_list();
		return;
	}

	if ((tp = fopen("tags", "r")) == NULL) {
		emsg("Can't open tags file");
		return;
	}

	while (fgets(lbuf, MAXTLIN, tp) != NULL) {
	
		if ((fname = (uchar *)strchr(lbuf, TAB)) == NULL) {
			emsg("Format error in tags file");
			return;
		}
		*fname++ = '\0';
		if ((str = (uchar *)strchr(fname, TAB)) == NULL) {
			emsg("Format error in tags file");
			return;
		}
		*str++ = '\0';

		if (strcmp(lbuf, tag) != 0)
			continue;

		fclose(tp);

		/*
		 * Scan through the search string. If we see a magic
		 * char, we have to quote it. This lets us use "real"
		 * implementations of ctags.
		 */
		p = pbuf;
		*p++ = *str++;		/* copy the '/' or '?' */
		*p++ = *str++;		/* copy the '^' */

		for (; *str != NUL ;str++) {
			if (*str == '\\') {
				*p++ = *str++;
				*p++ = *str;
			} else if (strchr("/?", *str) != NULL) {
				if (str[1] != '\n') {
					*p++ = '\\';
					*p++ = *str;
				} else
					*p++ = *str;
			} else if (strchr("*.", *str) != NULL) {
				*p++ = '\\';
				*p++ = *str;
			} else
				*p++ = *str;
		}
		*p = NUL;

		/*
		 * This looks out of order, but by calling stuffin()
		 * before doecmd() we keep an extra screen update
		 * from occuring. This stuffins() have no effect
		 * until we get back to the main loop, anyway.
		 */
		stuffin(pbuf);		/* str has \n at end */
		stuffin(mkstr(Ctrl('G')));

		cur_pos = tag_new();
		if (cur_pos == NULL || strcmp(cur_pos->file, fname) != 0) {
			if (doecmd(fname, force)) {  /* loaded file */
				tag_push(cur_pos);
			} else {	/* couldn't load file */
				stuffin(NULL);	/* clear the input */
				tag_dispose(cur_pos);
			}
		} else {	/* tag is within current file */
			tag_push(cur_pos);
		}
		return;
	}
	emsg("tag not found");
}

static	bool_t
doecmd(arg, force)
uchar	*arg;
bool_t	force;
{
	int	line = 1;		/* line # to go to in new file */

	if (!force && Changed) {
		emsg(nowrtmsg);
		if (altfile)
			free(altfile);
		altfile = strsave(arg);
		return FALSE;
	}
	if (arg != NULL) {
		/*
		 * First detect a ":e" on the current file. This is mainly
		 * for ":ta" commands where the destination is within the
		 * current file.
		 */
		if (Filename != NULL && strcmp(arg, Filename) == 0) {
			if (!Changed || (Changed && !force))
				return TRUE;
		}
		if (altfile) {
			if (strcmp (arg, altfile) == 0)
				line = altline;
			free(altfile);
		}
		altfile = Filename;
		altline = cntllines(Filemem, Curschar);
		Filename = strsave(arg);
	}
	if (Filename == NULL) {
		emsg("No filename");
		return FALSE;
	}

	/* clear mem and read file */
	freeall();
	filealloc();
	UNCHANGED;

	if (readfile(Filename, Filemem, 0))
		filemess("[New File]");

	*Topchar = *Curschar;
	if (line != 1) {
		stuffnum(line);
		stuffin("G");
	}
	do_mlines();
	setpcmark();
	updatescreen();
	return TRUE;
}

void
gotocmd(clr, firstc)
bool_t  clr;
uchar	firstc;
{
	windgoto(Rows-1,0);
	if (clr)
		toutstr(T_EL, 1);	/* clear the bottom line */
	if (firstc)
		outchar(firstc);
}

void
mmsg(s)
uchar	*s;
{
	windgoto(Rows-1, 0);
	toutstr(T_EL, 1);
	windgoto(Rows-1, Columns-20);
	outstr(s);
	flushbuf();
}

/*
 * msg(s) - displays the string 's' on the status line
 */
void
msg(s)
uchar	*s;
{
	gotocmd(TRUE, 0);
	if (standout) {
		if (T_SO) toutstr(T_SO, 1);
		outstr(s);
		if (T_SE) toutstr(T_SE, 1);
	} else
		outstr(s);
	flushbuf();
	if (strlen(s) >= Columns) {
		outstr("\r\n");
		wait_return();
		/*
		if (T_SO) toutstr(T_SO, 1);
		outstr("[Hit return to continue]");
		if (T_SE) toutstr(T_SE, 1);
		vgetc();
		gotocmd(TRUE, 0);
		flushbuf();
		*/
	}
}

/*VARARGS1*/
void
smsg(s, a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15,a16)
uchar	*s;
int	a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15,a16;
{
	uchar	sbuf[MAXMLEN];

	sprintf(sbuf, s,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15,a16);
	msg(sbuf);
}

/*
 * emsg() - display an error message
 *
 * Rings the bell, if appropriate, and calls message() to do the real work
 */
void
emsg(s)
uchar	*s;
{
	if (P(P_EB))
		beep();

	standout = TRUE;
	msg(s);
	standout = FALSE;
}

void
wait_return()
{
	register uchar	c;

	if (T_SO) toutstr(T_SO, 1);
	if (got_int)
		outstr("Interrupt: ");
	outstr("[Hit return to continue]");
	if (T_SE) toutstr(T_SE, 1);

	do {
		c = vgetc();
	} while (c != CR && c != NL && c != ' ' && c != ':');

	if (c == ':') {
		outchar(NL);
		docmdln(getcmdln(c));
	}
	screenclear();
	updatescreen();
}

/*
 *	Tag stack implementation - by Eric E. Bowles.
 *
 *	Tag_entry *tag_new()	- allocate a new tag entry.
 *	void tag_dispose(t)	- deallocate tag entry `t'.
 *	bool_t tag_push(t)	- save current position on tag stack.
 *	void tag_pop()		- return to most recent tag position.
 */
static Tag_entry *
tag_new()
{
	Tag_entry *t;

	if (Filename == NULL)
		return(NULL);
	if ((t = (Tag_entry *) alloc(sizeof(Tag_entry))) == NULL)
		return(NULL);
	if ((t->file = (char *)strsave(Filename)) == NULL) /* save file name */
		return(NULL);
	t->line = cntllines(Filemem, Curschar);		/* save line number */
	t->next = NULL;		/* for safety */
	return(t);
}

static void
tag_dispose(t)
	Tag_entry *t; 
{
	if (t == NULL)
		return;
	if (t->file)
		free((char *) t->file);
	free((char *) t);
}

static bool_t
tag_push(t)
	Tag_entry *t;
{
	if (t == NULL)
		return(FALSE);
	t->next = Tag_stack;		/* add the entry to the stack */
	Tag_stack = t;
	return(TRUE);
}

static void
tag_pop(force)
bool_t	force;
{
	Tag_entry *t;

	if (Tag_stack == NULL) {		/* no entries in the stack */
		emsg("Tagstack empty.");
		return;
	}
   	/* get the entry on top of the stack */
	t = Tag_stack;

	/* Edit new file if different from current one */
	if (strcmp(t->file, Filename) != 0) {
		if (Changed && !force) {
			/* file has been modified so we can't leave it */
			emsg(nowrtmsg);
			return;
		} 
		/* can load file immediately */
		stuffin(force? ":e! ": ":e ");
		stuffin((char *) t->file);
		stuffin("\n");
	}

	/* Go to the line: e.g., "40G z."  */
	stuffnum((int) t->line);
	stuffin("Gz.");
	Tag_stack = Tag_stack->next;	/* update top of stack */
	tag_dispose(t);
}

static void
tag_list()
{
	Tag_entry *t;
	char buf[MAXMLEN];

	if (Tag_stack == NULL) {		/* no entries in the stack */
		emsg("Tagstack empty.");
		return;
	}
	gotocmd(FALSE, 0);
	outstr("\r\n");
	for (t = Tag_stack; t != NULL; t = t->next) {
#ifdef V6FORMAT
		sprintf(buf, "%10D  %s\r\n", t->line, t->file);
#else
		sprintf(buf, "%10ld  %s\r\n", t->line, t->file);
#endif
		outstr(buf);
	}
	wait_return();
}

static void
showmap(type)
int	type;
{
	register MAP	*m;
	int	i, col;

	gotocmd(FALSE, 0);
	outstr("\r\n");
	for (m = Maptab, i = 0; i < nmap; i++, m++) {
		if (m->m_type != type)
			continue;
		col = 0;
		outstr(pretty((uchar *)m->m_name, &col));
		outstr(pretty(mkstr(TAB), &col));
		outstr(pretty((uchar *)m->m_key, &col));
		outstr(pretty(mkstr(TAB), &col));
		outstr(pretty((uchar *)m->m_str, &col));
		outstr("\r\n");
	}
	wait_return();
}

/*
 * domap - process :map, :map!, :ab command
 */
void
domap(arg, type, set)
uchar	*arg;
int	type;
bool_t	set;
{
	register uchar	*p, *q;
	int	i;
	MAP	*m;
	uchar	buf[MAXCLIN];

	if (arg == NULL) {
		if (set == TRUE) {
			showmap(type);
			return;
		}
		emsg("Missing lhs");
		return;
	}
	if (set == FALSE) {
		for (m = Maptab, i = 0; i < nmap; i++, m++) {
			if (m->m_type != type)
				continue;
			if (strcmp(m->m_key, arg) == 0)
				break;
		}
		if (i == nmap) {
			emsg("That macro wasn't mapped");
			return;
		}
		free((char *)m->m_name);
		free((char *)m->m_key);
		free((char *)m->m_str);
		for (; i < nmap - 1; i++)
			m[0] = m[1];
		nmap--;
		return;
	}
	if (nmap == MAXMAP) {
		emsg("Too many macros");
		return;
	}
	p = arg;
	while (*p != NUL && *p != ' ' && *p != TAB)
		p++;
	q = p;
	while (*p != NUL && (*p == ' ' || *p == TAB))
		p++;
	if (*p == NUL) {
		emsg("Missing rhs");
		return;
	}
	*q = NUL;
	if (strcmp(arg, p + strlen(p) - strlen(arg)) == 0) {
		emsg("No tail recursion");
		return;
	}
	for (m = Maptab, i = 0; i < nmap; i++, m++) {
		if (m->m_type != type)
			continue;
		if (strcmp(m->m_key, arg) == 0)
			break;
	}
	if (i == nmap) {
		nmap++;
	} else {
		free((char *)m->m_name);
		free((char *)m->m_key);
		free((char *)m->m_str);
	}
	m->m_name = (char *)strsave(arg);
	m->m_key = (char *)strsave(arg);
	m->m_str = (char *)strsave(p);
	m->m_type = type;
	m->m_len = strlen(m->m_key);
}
