/* $Header: /nw/tony/src/stevie/src/RCS/dos.c,v 1.7 89/07/11 21:22:01 tony Exp $
 *
 * DOS System-dependent routines.
 *
 * System-specific code for MS-DOS. This has been tested with
 * MSDOS 3.3 on an AT. Also, the console driver "nansi.sys" is
 * required.
 */

#include "stevie.h"
#ifdef __TURBOC__
#define beep tcbeep
#include <dos.h>
#undef beep
#else
#include <time.h>
#include <dos.h>
#include <sys/types.h>
#endif
#include <sys/stat.h>
#include <signal.h>
#include <process.h>

#ifdef PC98_MOUSE
#include "pc98_mou.h"
#endif

#define	getch()		(bdos(7, 0x00, 0) & 0xff)
#define	getchnowait()	(bdos(6, 0xff, 0) & 0xff)

uchar	*Initfile = "stevie.rc";

static	uchar	getswitch();
static	void	setswitch();

#ifndef PC98_VRAM

int	OutInt, OutFunc;

#ifdef PC98_50LINE
#include "pc98_50l.inc"
#endif

/*
 * inchar() - get a character from the keyboard
 */
int
inchar()
{
	int	c;

	got_int = FALSE;

	for (;;beep()) {	/* loop until we get a valid character */

		flushbuf();	/* flush any pending output */

#ifdef PC98_50LINE
		if (cursor.should ^ cursor.is)
			drawcursor();
#ifdef FEPCTRL
		code_access_kcg();
#endif
#endif
#ifdef KANJI
		c = getch();
		if (isKanji(c))
			c = (c << 8) | getch();
#ifdef PC98_50LINE
#ifdef FEPCTRL
		dot_access_kcg();
#endif
#endif
		return c;
#else /*!KANJI*/
		switch (c = getch()) {
		case 0x1e:
			return K_CGRAVE;
		case 0:				/* special key */
			if (State != NORMAL) {
				c = getch();	/* throw away next char */
				continue;	/* and loop for another char */
			}
			switch (c = getch()) {
			case 0x50:
				return K_DARROW;
			case 0x48:
				return K_UARROW;
			case 0x4b:
				return K_LARROW;
			case 0x4d:
				return K_RARROW;
			case 0x52:
				return K_INSERT;
			case 0x47:
				stuffin("1G");
				return -1;
			case 0x4f:
				stuffin("G");
				return -1;
			case 0x51:
				stuffin(mkstr(Ctrl('F')));
				return -1;
			case 0x49:
				stuffin(mkstr(Ctrl('B')));
				return -1;
			/*
			 * Hard-code some useful function key macros.
			 */
			case 0x3b: /* F1 */
				stuffin(":N\n");
				return -1;
			case 0x54: /* SF1 */
				stuffin(":N!\n");
				return -1;
			case 0x3c: /* F2 */
				stuffin(":n\n");
				return -1;
			case 0x55: /* SF2 */
				stuffin(":n!\n");
				return -1;
			case 0x3d: /* F3 */
				return K_CGRAVE;
			case 0x3e: /* F4 */
				stuffin(":rew\n");
				return -1;
			case 0x57: /* SF4 */
				stuffin(":rew!\n");
				return -1;
			case 0x3f: /* F5 */
				stuffin("[[");
				return -1;
			case 0x40: /* F6 */
				stuffin("]]");
				return -1;
			case 0x41: /* F7 - explain C declaration */
				stuffin("yyp^iexplain \033!!cdecl\n");
				return -1;
			case 0x42: /* F8 - declare C variable */
				stuffin("yyp!!cdecl\n");
				return -1;
			case 0x43: /* F9 */
				stuffin(":x\n");
				return -1;
			case 0x44: /* F10 */
				stuffin(":help\n");
				return -1;
			default:
				break;
			}
			break;
		default:
			return c;
		}
#endif
	}
}

#ifdef PC98_50LINE
static	uchar	outbuf[OBSIZE+1];
#else
static	uchar	outbuf[OBSIZE];
#endif
static	int	bpos = 0;

void
flushbuf()
{
#ifdef PC98_50LINE
	if (bpos == 0)
		return;

	outbuf[bpos] = '\0';
	prnstr( outbuf );
#else
	register int n;
	register uchar *p;
#ifndef __TURBOC__
	union REGS regs;
#endif

	if (bpos == 0)
		return;

	n = bpos;
	p = outbuf;
	while (n--) {
#ifdef __TURBOC__
		_AH = OutFunc;
		_AL = _DL = *p++;
		geninterrupt(OutInt);
#else
		regs.h.ah = OutFunc;
		regs.h.al = regs.h.dl = *p++;
		int86(OutInt, &regs, &regs);
#endif
	}
#endif
	bpos = 0;
}

/*
 * Macro to output a character. Used within this file for speed.
 */
#define	outone(c)	outbuf[bpos++] = c; if (bpos >= OBSIZE) flushbuf()

/*
 * Function version for use outside this file.
 */
void
outchar(c)
uchar	c;
{
	outone(c);
}

/*
 * outstr(s) - write a string to the console
 */
void
outstr(s)
register uchar	*s;
{
	while (*s) {
		outone(*s++);
	}
}

void
beep()
{
	outone('\007');
}

#else /*PC98_VRAM*/

#define	A_NORM	0xe1		/* normal white */
#define	A_REV	0xa5		/* reverse cyan */
#define	A_UL	0xc9		/* underline yellow */

static	int	Currow = 0, Curcol = 0;
static	int	Savrow, Savcol;
static	uchar	Curattr = A_NORM;

#define	FAR(s,o) (unsigned char far *)((unsigned long)(s)<<16|(unsigned)(o))
static	unsigned short far *vram  = (unsigned short far *)FAR(0xa000, 0);
static	unsigned short far *vattr = (unsigned short far *)FAR(0xa200, 0);

/*
 * inchar() - get a character from the keyboard
 */
int
inchar()
{
	int	c;
	bool_t	busy;
#ifdef PC98_MOUSE
	int	x, y, x0, y0, x1, y1, xx, yy;
	char	attr1, attr2;
	union	event_t	event;
	static	short	yanklen;
	static	char	yankbuf[YANK_MAX+1];
	static	int	old_ai;
	static	bool_t	ai_saved = FALSE;
#endif

	got_int = FALSE;
	busy = anyinput();
	if (!busy)
		dispctl(T_CV);
#ifdef PC98_MOUSE
	NextEvent(&event);
	if (event.key.type == KEYINPUT) {
		if (ai_saved) {
			P(P_AI) = old_ai;
			ai_saved = FALSE;
		}
		c = event.key.code;
		if (isKanji(c))
			c = (c << 8) | getch();
	} else if (event.key.type == MOUSE_DOWN_L) {
		PosToCol(event.mouse.x, event.mouse.y, &x0, &y0, 1);
		AttrGet(x0, y0, &attr1);
		AttrSet(x0, y0, ATTR_REV);
		x1 = y1 = -1;
		xx = x0;
		yy = y0;
		while (!MouseReleaseL(&x, &y)) {
			MouseMove(&x, &y);
			PosToCol(x, y, &x1, &y1, 0);
			if (!(x0==xx && y0==yy) && !(x1==xx && y1==yy))
				AttrSet(xx, yy, attr2);
			if (!(x0==x1 && y0==y1) && !(x1==xx && y1==yy)) {
				AttrGet(x1, y1, &attr2);
				AttrSet(x1, y1, ATTR_REV);
			}
			if (!(x1 == xx && y1 == yy)) {
				xx = x1;
				yy = y1;
			}
		}
		AttrSet(x0, y0, attr1);
		if (!(x0 == x1 && y0 == y1))
			AttrSet(x1, y1, attr2);
		PosToCol(x, y, &x1, &y1, 0);
		VramToString(x0, y0, x1, y1, yankbuf, &yanklen);
		c = -1;
	} else if (event.key.type == MOUSE_DOWN_R) {
		while (!MouseReleaseR(&x, &y))
			;
		if (State != NORMAL) {
			old_ai = P(P_AI);
			P(P_AI) = FALSE;
			ai_saved = TRUE;
			stuffin(yankbuf);
		}
		c = -1;
	}
#else
	c = getch();
	if (isKanji(c))
		c = (c << 8) | getch();
#endif
	if (!busy)
		dispctl(T_CI);
	return c;
}

void
flushbuf()
{
	return;
}

/*
 * Function version for use outside this file.
 */
void
outchar(c)
register unsigned short	c;
{
	register unsigned short	i;
	unsigned short	endpos;
	static	uchar	k1 = 0;
	extern	uchar	stoj1[], stoj2[];

	if (k1) {
		i = (Currow * Columns) + Curcol;
		Curcol += 2;
		vram[i] = c = (stoj2[c] << 8) + stoj1[k1] + (c >= 0x9f);
		vram[i+1] = c | 0x8000;
		*(uchar far *)&vattr[i] = *(uchar far *)&vattr[i+1] = Curattr;
		if (Curcol >= Columns) {
			Currow++;
			Curcol = 0;
		}
		k1 = 0;
	} else if (isKanji(c)) {
		k1 = c;
	} else if (c == CR) {
		Curcol = 0;
	} else if (c == NL) {
		Currow++;
	} else {
		i = (Currow * Columns) + Curcol++;
		vram[i] = c;
		*(uchar far *)&vattr[i] = Curattr;
		if (Curcol >= Columns) {
			Currow++;
			Curcol = 0;
		}
	}
	if (Currow >= Rows) {
		endpos = Rows * Columns;
		i = 0;
		c = Columns;
		while (c < endpos)
			 vram[i++] = vram[c++];
		i = (Rows - 1) * Columns;
		while (i < endpos)
			vram[i++] = ' ';
		Currow--;
	}
}

/*
 * outstr(s) - write a string to the console
 */
void
outstr(s)
register uchar	*s;
{
	while (*s)
		outchar(*s++);
}

void
putch(c)
int	c;
{
#ifdef __TURBOC__
	_AH = 2;
	_AL = c;
	geninterrupt(0x29);
#else
	union REGS regs;
	regs.h.ah = 2;
	regs.h.al = c;
	int86(0x29, &regs, &regs);
#endif
}

void
beep()
{
	putch('\007');
}

#endif /*PC98_VRAM*/

#ifndef __TURBOC__
sleep(n)
int	n;
{
	time_t	t;

	t = time((time_t *)0);
	while (time((time_t *)0) <= t + n)
		/* void */;
}
#endif

void
delay()
{
	long	l;

	flushbuf();
#ifdef PC98_50LINE
	if (cursor.should ^ cursor.is)
		drawcursor();
#endif
	/*
	 * Should do something better here...
	 */
	for (l=0; l < 30000 ;l++)
		;
}

#if defined(__TURBOC__) && __TURBOC__ < 0x0200
int
sig()
{
	ctrlbrk(sig);
	got_int = TRUE;
}
#else
void
sig()
{
	signal(SIGINT, sig);
	got_int = TRUE;
}
#endif

void
tstpon()
{
	/* nothing to do here */
}

void
tstpoff()
{
	/* nothing to do here */
}

static	uchar	schar;		/* save original switch character */

void
windinit()
{
#ifdef	TERMCAP
	switch (t_init()) {
	case 1:
		break;
	case 0:
		fprintf(stderr, "insufficient terminal capability\n");
		exit(1);
	case -1:
		Columns = 80;
		P(P_LI) = Rows = 24;
		T_CM = "\033[%i%d;%dH";
		T_EB = "\033[J";
		T_EL = "\033[K";
		T_IL = "\033[L";
		T_DL = "\033[M";
		T_ED = "\033[2J";
		T_SC = "\033[s";
		T_RC = "\033[u";
		T_SO = "\033[7m";
		T_SE = "\033[m";
		T_US = "\033[4m";
		T_UE = "\033[m";
		T_CI = NULL;
		T_CV = NULL;
		T_TI = NULL;
		T_TE = NULL;
		break;
	}
#else
	Columns = 80;
#ifdef PC98_50LINE
#ifdef FEPCTRL
	P(P_LI) = Rows = 48; /* set lines */
#else
	P(P_LI) = Rows = 50; /* set lines */
#endif
	P(P_SS) = Rows / 2; /* set scroll length */
#else
	P(P_LI) = Rows = 24;
#endif
#endif

	schar = getswitch();
	setswitch('/');

#if defined(__TURBOC__) && __TURBOC__ < 0x0200
	ctrlbrk(sig);
#else
	signal(SIGINT, sig);
#endif
#ifdef PC98_VRAM
	dispctl(T_CI);
	rawmode(1);
#ifdef FEPCTRL
	if (P(P_FC))
		fep_init();
#endif
#ifdef PC98_MOUSE
	mouse_init();
#endif
#else /*PC98_VRAM*/
	setoutfunc();
	rawmode(1);
#ifdef FEPCTRL
	if (P(P_FC))
		fep_init();
#endif
#ifdef PC98_50LINE
	openscreen();
#else
	if (T_TI) toutstr(T_TI, 1);
#endif
#endif /*PC98_VRAM*/
}

void
windexit(r)
int r;
{
#ifdef PC98_VRAM
#ifdef PC98_MOUSE
	mouse_exit();
#endif
	dispctl(T_CV);
	dispmove(Currow, Curcol);
#else /*PC98_VRAM*/
#ifdef PC98_50LINE
	flushbuf();
	closescreen();
#else
	if (T_TE) toutstr(T_TE, 1);
	flushbuf();
#endif
#endif /*PC98_VRAM*/
#ifdef FEPCTRL
	if (P(P_FC))
		fep_term();
#endif
	rawmode(0);
	setswitch(schar);
	exit(r);
}

void
windgoto(r, c)
register int	r, c;
{
#ifdef PC98_VRAM
	Currow = r;
	Curcol = c;
#else
#ifdef PC98_50LINE
	if (cursor.is)
		drawcursor();
	flushbuf();
	cursor.x = c; cursor.y = r; ADJCURSOR();
#else
#ifdef	TERMCAP
	char	*tgoto();

	toutstr(tgoto(T_CM, c, r), 1);
#else
	r++;
	c++;

	/*
	 * Check for overflow once, to save time.
	 */
	if (bpos + 8 >= OBSIZE)
		flushbuf();

	outbuf[bpos++] = '\033';
	outbuf[bpos++] = '[';
	if (r >= 10) {
		outbuf[bpos++] = r / 10 + '0';
		outbuf[bpos++] = r % 10 + '0';
	} else
		outbuf[bpos++] = r + '0';
	outbuf[bpos++] = ';';
	if (c >= 10) {
		outbuf[bpos++] = c / 10 + '0';
		outbuf[bpos++] = c % 10 + '0';
	} else
		outbuf[bpos++] = c + '0';
	outbuf[bpos++] = 'H';
#endif
#endif
#endif
}

#ifdef PC98_VRAM
void
dispmove(r, c)
register int	r, c;
{
	putch(ESC);
	putch('=');
	putch(' ' + r);
	putch(' ' + c);
}

void
dispctl(c)
int	c;
{
	register int	pos, endpos;

	switch (c) {
	case T_EL:
		pos = Currow * Columns + Curcol;
		endpos = (Currow + 1) * Columns;
		while (pos < endpos) {
			vram[pos] = ' ';
			*(uchar far *)&vattr[pos++] = A_NORM;
		}
		break;
	case T_ED:
		pos = 0;
		endpos = Rows * Columns;
		while (pos < endpos) {
			vram[pos] = ' ';
			*(uchar far *)&vattr[pos++] = A_NORM;
		}
		Currow = Curcol = 0;
		break;
	case T_SC:
		Savrow = Currow;
		Savcol = Curcol;
		break;
	case T_RC:
		Currow = Savrow;
		Curcol = Savcol;
		break;
	case T_SO:
		Curattr = A_REV;
		break;
	case T_US:
		Curattr = A_UL;
		break;
	case T_SE:
	case T_UE:
		Curattr = A_NORM;
		break;
	case T_CV:
		dispmove(Currow, Curcol);
		/* fall thru ... */
	case T_CI:
		putch(ESC);
		putch('[');
		putch('>');
		putch('5');
		putch((c == T_CI)? 'h': 'l');
		break;
	default:
		break;
	}
}
#endif /*PC98_VRAM*/

static	uchar
getswitch()
{
	union	REGS	inregs, outregs;

	inregs.h.ah = 0x37;
	inregs.h.al = 0;

	intdos(&inregs, &outregs);

	return outregs.h.dl;
}

static	void
setswitch(c)
uchar	c;
{
	union	REGS	inregs, outregs;

	inregs.h.ah = 0x37;
	inregs.h.al = 1;
	inregs.h.dl = c;

	intdos(&inregs, &outregs);
}

FILE *
fopenb(fname, mode)
uchar	*fname;
uchar	*mode;
{
	FILE	*fopen();
	uchar	modestr[16];

	sprintf(modestr, "%sb", mode);
	return fopen(fname, modestr);
}

int
filestat(fname)
uchar	*fname;
{
	int	ret;
	struct stat stbuf;

	if (stat(fname, &stbuf) < 0)
		return 0;

	ret = FS_EXIST;
	if (stbuf.st_mode & S_IREAD)
		ret |= FS_READ;
	if (stbuf.st_mode & S_IWRITE)
		ret |= FS_WRITE;
	if (stbuf.st_mode & S_IFDIR)
		ret |= FS_DIR;

	return ret;
}

/*
 * Rename(of, nf) - rename existing file 'of' to 'nf'
 */
int
Rename(of, nf)
uchar	*of, *nf;
{
	unlink(nf);
	return rename(of, nf);
}

/*
 * backupname(fname, ext) - generate filename for backup
 */
uchar *
backupname(fname, ext)
uchar	*fname, *ext;
{
	register uchar	*s;
	static uchar	bname[MAXPLEN];

	strcpy(bname, fname);
	for (s = bname; *s; s++)
		/* void */;
	while (s > bname && *s != '/' && *s != '\\' && *s != ':')
		s--;
	while (*s && *s != '.')
		s++;
	*s = NUL;
	strcat(bname, ext);
	return bname;
}

/*
 * fixname(s) - fix up a dos name
 *
 * Takes a name like:
 *
 *	\x\y\z\base.ext
 *
 * and trims 'base' to 8 characters, and 'ext' to 3.
 */
uchar *
fixname(s)
uchar	*s;
{
	static	uchar	f[MAXPLEN];
	uchar	base[32];
	uchar	ext[32];
	uchar	*p;
	int	i;

	strcpy(f, s);

	for (i=0; i < MAXPLEN ;i++)
		if (f[i] == '/')
			f[i] = '\\';

	/*
	 * Split the name into directory, base, extension.
	 */
	if ((p = strrchr(f, '\\')) != NULL || (p = strrchr(f, ':')) != NULL) {
		strcpy(base, p+1);
		p[1] = '\0';
	} else {
		strcpy(base, f);
		f[0] = '\0';
	}

	if ((p = strchr(base, '.')) != NULL) {
		strcpy(ext, p+1);
		*p = '\0';
	} else
		ext[0] = '\0';

	/*
	 * Trim the base name if necessary.
	 */
	if (strlen(base) > 8)
		base[8] = '\0';
	
	if (strlen(ext) > 3)
		ext[3] = '\0';

	/*
	 * Paste it all back together
	 */
	strcat(f, base);
	strcat(f, ".");
	strcat(f, ext);

	return f;
}

uchar *
mkpath(dir, file)
uchar	*dir, *file;
{
	int	n;
	static uchar	buf[MAXPLEN];

	n = 0;
	if (dir) {
		strcpy(buf, dir);
		n = strlen(buf);
		if (buf[n-1] != '/' && buf[n-1] != '\\')
			buf[n++] = '\\';
	}
	strcpy(buf+n, file);
	return buf;
}

void
doshell(cmd)
uchar	*cmd;
{
	extern char *getenv(const char *);

	outstr("\r\n");
	flushbuf();
#ifdef PC98_50LINE
	stopscreen(); /* stop graphics and switch to text screen */
#endif
#ifdef FEPCTRL
	if (P(P_FC))
		fep_term();
#endif
#ifdef PC98_VRAM
#ifdef PC98_MOUSE
	mouse_exit();
#endif
	dispmove(++Currow, Curcol);
	dispctl(T_CV);
#endif
	rawmode(0);
	if (cmd){
		system(cmd);
	} else {
		cmd = getenv( "COMSPEC" );
		if (cmd == NULL)
			cmd = "command.com";
		spawnl( P_WAIT, cmd, cmd, NULL );
	}
	rawmode(1);
#ifdef PC98_VRAM
	dispmove(Currow = Rows - 1, Curcol = 0);
	dispctl(T_CI);
#ifdef PC98_MOUSE
	mouse_init();
#endif
#endif
#ifdef FEPCTRL
	if (P(P_FC))
		fep_init();
#endif
#ifdef PC98_50LINE
	/*
	 * here are codes took from wait_return() in cmdline.c.
	 * PC98_50LINE need special version of wait_return() here
	 * as it will execute system() in normal screen
	 */
	if (T_SO) toutstr(T_SO, 1);
	if (got_int)
		printf("Interrupt: ");
	printf("[Hit return to continue]");
	if (T_SE) toutstr(T_SE, 1);
	(void)vgetc();
	putchar( '\n' ); /* return text cursor to top of line */
	startscreen(); /* stop text and return to graphics screen */
	screenclear();
	updatescreen();
#else
	wait_return();
#endif
}

void
cmdread(cmd)
uchar	*cmd;
{
	int fd, old, err;
	char tmpfile[MAXPLEN];

	if (cmd[0] == NUL) {
		emsg("Exec failed");
		return;
	}
	tmpnam(tmpfile);
	if ((fd = creat(tmpfile, 0600)) < 0) {
		emsg("Can't create temporary file");
		return;
	}
	old = dup(1);
	dup2(fd, 1);
	close(fd);
#ifdef PC98_VRAM
	dispmove(Currow, Curcol = 0);
	dispctl(T_CV);
	err = system(cmd);
	dispctl(T_CI);
#else
	outstr("\r");
	flushbuf();
	err = system(cmd);
#endif
	close(1);
	dup2(old, 1);
	if (err < 0)
		emsg("Exec failed");
	else {
		if (readfile(tmpfile, Curschar, 1))
			emsg("Can't open temporary file");
		else {
			updatescreen();
			CHANGED;
		}
	}
	remove(tmpfile);
}

#define	GETCBRK	0x3300
#define	SETCBRK	0x3301
#define	GETVECT	0x35
#define IOCTL	0x44
#define DEVICE	0x80
#define RAW	0x20
#define SPDEV	0x10
#define GETBITS	0
#define SETBITS	1
#define	GETDATA	2

static unsigned
ioctl(handle, mode, setvalue)
unsigned setvalue;
{
	union REGS regs;

	regs.h.ah = IOCTL;
	regs.h.al = mode;
	regs.x.bx = handle;
	regs.h.dl = setvalue;
	regs.h.dh = 0;
	intdos(&regs, &regs);
	return ((unsigned)regs.x.dx);
}

static int
rawmode(on)
int on;
{
	static unsigned old0, old1;

	if (on) {
		old0 = ioctl(0, GETBITS, 0);
		if (old0 & DEVICE)
			ioctl(0, SETBITS, (old0 | RAW));
		old1 = ioctl(1, GETBITS, 0);
		if (old1 & DEVICE)
			ioctl(1, SETBITS, (old1 | RAW));
	} else {
		ioctl(0, SETBITS, (old0 & ~RAW));
		ioctl(1, SETBITS, (old1 & ~RAW));
	}
}

#ifndef PC98_VRAM
static int
setoutfunc()
{
	int dc;

	dc = ioctl(1, GETBITS, 0);
	if (dc & SPDEV) {
		OutInt = 0x29;
		OutFunc = 0x02;
	} else {
		OutInt = 0x21;
		OutFunc = 0x06;
	}
}
#endif /*!PC98_VRAM*/
