/*  >>> MODIFIED FOR USE WITH JSTEVIE <<<
 *
 *  This glob routine requires the UNIX directory(3) functions (opendir,
 *  readdir, closedir).  If your system isn't UNIX and it doesn't have these 
 *  functions, then try to find a public domain implementation (or write your 
 *  own).
 *
 *  Most of the modifications in this file involved adding #ifdefs to 
 *  accomodate various operating systems.  Also added fixstr() to force 
 *  uppercase names and change `/` to `\\` in DOS.  Modified call to
 *  opendir() so that an empty path would open the *current directory*.
 *
 *  Note: [...] IS supported, contrary to what it says below.  However,
 *  {...} is not implemented.  
 *                                                             -- Eric B.
 */

/* This software is copyright (c) 1986 by Stichting Mathematisch Centrum.
 * You may use, modify and copy it, provided this notice is present in all
 * copies, modified or unmodified, and provided you don't make money for it.
 *
 * Written 86-jun-28 by Guido van Rossum, CWI, Amsterdam <guido@mcvax.uucp>
 */

/*
 * 'Glob' routine -- match * and ? in filename.
 * Extra services: initial ~username is replaced by username's home dir,
 * initial ~ is replaced by $HOME, initial $var is replaced by
 * return value of getenv("var").
 * Quoting convention: \ followed by a magic character inhibits its
 * special meaning.
 * BUGS: [...] is not implemented.
 * Nonexisting $var is treated as empty string; nonexisting ~user
 * is copied literally.
 */

#include <ctype.h>
#include <stdio.h>

#define STEVIE
#ifdef STEVIE
#   include "env.h"     /* figure out which system we're running on */
#endif

#if defined(PC98_VRAM) || defined(PC98_50LINE)
#define	DOS
#endif

#ifdef UNIX
#include <sys/types.h>
#endif

#ifdef BSD
#   include <strings.h>
#else
#   include <string.h>
#   define index(s, c)   strchr(s, c)
#endif


/*---------------------------------------------------------------*/
#if defined(STEVIE) && !defined(GLOB)  /* Use dummy glob routine */
/*---------------------------------------------------------------*/
int glob(pat, buf, size)
    char *pat, *buf;
    unsigned int size; /* sizeof buf */
{
    strncpy(buf, pat, size);
    return(0);
}
/*---------------------------------------------------------------*/
#else   /* using glob() ---- continue with the rest of this file */
/*---------------------------------------------------------------*/


#ifdef UNIX
#   include <pwd.h>
#endif

/*
 *  Include appropriate directory header file.  
 *
 *  UNIX users:  Systems up to 4.3BSD use <sys/dir.h> and `struct direct'.
 *  4.4BSD is supposed to use <dirent.h> and `struct dirent', which is 
 *  supported by POSIX.
 *  Most System Vs since Release 3.0 also use <dirent.h> and `struct dirent'.  
 *  SunOs 4.0 encourages <dirent.h> and `struct dirent', although they have 
 *  <sys/dir.h> for backward compatibility (I know that SunOS 2.0 used it).  
 *
 *  If you don't have the directory routines like opendir, readdir, and 
 *  closedir, then you probably have an old UNIX (or a non-standard version); 
 *  get Douglas A. Gwyn's public domain POSIX-compliant library (send email 
 *  to <Gwyn@BRL.MIL>).  
 *                                                            -- Eric B.
 */

/*----------*/
/* Temporary definitions for Jstevie 1.3 -- will remove in the future. */
#if defined(UNIX) && !defined(HAS_SYSDIR) && !defined(HAS_DIRENT)
#   ifdef BSD
#       define HAS_SYSDIR       /* 4.2BSD and 4.3BSD have <sys/dir.h> */
#   else
#       define HAS_DIRENT       /* SVR3.0 or higher have <dirent.h> */
#   endif
#endif
/*----------*/


/* No #elif's used since we have to accomodate old preprocessors */
#ifdef UNIX
#   ifdef HAS_DIRENT
#       include <dirent.h>
#   else
#       ifdef HAS_SYSDIR
#           include <sys/dir.h>     
#       else
#           include "*** ERROR: Can't find directory header file ***"
#       endif
#   endif
#endif

#if defined(DOS) || defined(OS2)
#   include "dir.h"
#endif

/* Handle all other cases here -- need to modify as we add more systems */
#if ! ( defined(UNIX) || defined(DOS) || defined(OS2) )
#   include "*** ERROR: Globbing is not implemented for your system ***"
#endif


#ifdef UNIX
    struct passwd *getpwnam();
#endif

#if defined(HAS_SYSDIR) || defined(DOS) || defined(OS2)
typedef struct direct  ENTRY;   
#else
typedef struct dirent  ENTRY;   
#endif

char *getenv();
char *malloc();
char *realloc();

#define EOS '\0'

#if defined(DOS) || defined(OS2)
#   define SEP '\\'
#else
#   define SEP '/'
#endif

#define DOT '.'

#if defined(DOS) || defined(OS2)
#   define QUOTE '\''   /* a hack since we already use backslash for SEP */
#else
#   define QUOTE '\\'
#endif

#define BOOL int
#define NO 0
#define YES 1

#define MAXPATH 1024
#define MAXBASE 256

#define META(c) ((char)((c)|0200))
#define M_ALL META('*')
#define M_ONE META('?')
#define M_SET META('[')
#define M_RNG META('-')
#define M_END META(']')

struct flist {
    int len;
    char **item;
};

#if defined(DOS) || defined(OS2)
/* 
 *  Force uppercase letters and convert / to \ .
 */
static void
fixstr(str)
    char *str;
{
    char *p;

    for (p = str;  *p != EOS;  p++) {
        if (islower(*p))
            *p = toupper(*p);
        else if (*p == '/')
            *p = SEP;
    }
}
#endif /* DOS || OS2 */


/* Make a null-terminated string of the chars between two pointers */
/* (Limited length, static buffer returned) */

static char *
makestr(start, end)
    char *start;
    char *end;
{
    static char buf[100];
    char *p= buf;
    
    while (start < end && p < &buf[100])
        *p++ = *start++;
    *p= EOS;
#if defined(DOS) || defined(OS2)
    fixstr(buf);
#endif
    return buf;
}

/* Append string to buffer, return new end of buffer.  Guarded. */

static char *
addstr(dest, src, end)
    char *dest;
    char *src;
    char *end;
{
#if defined(DOS) || defined(OS2)
    char *start = dest;
#endif

    while (*dest++ = *src++) {
        if (dest >= end)
            break;
    }
#if defined(DOS) || defined(OS2)
    fixstr(start);
#endif
    return --dest;
}

/* Form a pathname by concatenating head, maybe a / and tail. */
/* Truncates to space available. */

static void
formpath(dest, head, tail, size)
    char *dest;
    char *head;
    char *tail;
    unsigned int size; /* sizeof dest */
{
    char *start= dest;
    
    for (;;) {
        if (--size == 0)
            goto out;
        if ((*dest++ = *head++) == EOS)
            break;
    }
    if (*tail != EOS && size != 0) {
        --dest;
        ++size;
        if (dest > start && dest[-1] != SEP) {
            *dest++ = SEP;
            --size;
        }
        for (;;) {
            if (--size == 0)
                goto out;
            if ((*dest++ = *tail++) == EOS)
                break;
        }
    }
    return;
    
 out:   *dest= EOS;
}

/* Find a user's home directory, NULL if not found */

static char *
gethome(username)
    char *username;
{
#ifdef UNIX
    struct passwd *p;
    
    p= getpwnam(username);
    if (p == NULL)
        return NULL;
    return p->pw_dir;
#else
    return( NULL );
#endif
}

/* String compare for qsort */

static int
compare(p, q)
    char **p;
    char **q;
{
    return strcmp(*p, *q);
}

/*
 * Maintain lists of strings.
 * When memory is full, data is lost but 
 */

static void
addfile(list, head, tail)
    struct flist *list;
    char *head;
    char *tail;
{
    char *str;
    
    str= malloc((unsigned) strlen(head) + strlen(tail) + 2);
    
    if (str == NULL)
        return;
    if (list->item == 0) {
        list->len= 0;
        list->item= (char**) malloc(sizeof(char*));
    }
    else {
        list->item= (char**) realloc((char*)list->item,
                (unsigned) (list->len+1)*sizeof(char*));
        if (list->item == 0)
            list->len= 0;
    }
    if (list->item != NULL) {
        formpath(str, head, tail, MAXPATH);
        list->item[list->len++]= str;
    }
    else
        free(str);
}

/* Clear the fields of a struct flist before first use */

static void
clear(list)
    struct flist *list;
{
    list->len= 0;
    list->item= NULL;
}

/* Free memory held by struct flist after use */

static void
discard(list)
    struct flist *list;
{
    int i= list->len;
    
    if (list->item != NULL) {
        while (--i >= 0) {
            if (list->item[i] != NULL)
                free(list->item[i]);
        }
        free((char*)list->item);
        list->item= NULL;
    }
    list->len= 0;
}

/* Pattern matching function for filenames */
/* Each occurrence of the * pattern causes a recursion level */

static BOOL
match(name, pat)
    char *name;
    char *pat;
{
    char c, k;
    BOOL ok;
    
    while ((c= *pat++) != EOS) {
        switch (c) {

        case M_ONE:
            if (*name++ == EOS)
                return NO;
            break;

        case M_ALL:
            if (*pat == EOS)
                return YES;
            for (; *name != EOS; ++name) {
                if (match(name, pat))
                    return YES;
            }
            return NO;

        case M_SET:
            ok= NO;
            k= *name++;
            while ((c= *pat++) != M_END) {
                if (*pat == M_RNG) {
                    if (c <= k && k <= pat[1])
                        ok= YES;
                    pat += 2;
                }
                else if (c == k)
                    ok= YES;
            }
            if (!ok)
                return NO;
            break;

        default:
            if (*name++ != c)
                return NO;
            break;

        }
    }
    return *name == EOS;
}

/* Perform pattern matching for one segment of the pathname */

static void
segment(files, mid, pat)
    struct flist *files;
    char *mid;
    char *pat;
{
    char path[MAXPATH];
    int i;
    DIR *dirp;
    ENTRY *dp;
    struct flist new;
    
    clear(&new);
    for (i= 0; i < files->len; ++i) {
        strcpy(path, files->item[i]);
        strcat(path, mid);
        free(files->item[i]);
        files->item[i]= NULL;
        dirp= opendir( (*path == NULL) ? "." : path );
        if (dirp == NULL)
            continue;
        while ((dp= readdir(dirp)) != NULL) {
            if (*dp->d_name == DOT && *pat != DOT)
                ; /* No meta-match on initial '.' */
            else if (match(dp->d_name, pat))
                addfile(&new, path, dp->d_name);
        }
        closedir(dirp);
    }
    discard(files);
    *files= new;
}

/* Finish by matching the rest of the pattern (which has no metas) */

static void
findfiles(files, tail)
    struct flist *files;
    char *tail;
{
    int i;
    struct flist new;
    char path[MAXPATH];

    if (*tail == EOS || files->len == 0)
        return;
    clear(&new);
    for (i= 0; i < files->len; ++i) {
        strcpy(path, files->item[i]);
        strcat(path, tail);
        free(files->item[i]);
        files->item[i]= NULL;
#ifdef UNIX
        if (access(path, 0) == 0)
#endif
            addfile(&new, path, "");
    }
    discard(files);
    *files= new;
}

static void
glob1(pat, files)
    char *pat;
    struct flist *files;
{
    char mid[MAXPATH];
    char *end= mid;
    char *basestart= mid;
    BOOL meta= NO;
    char c;

    clear(files);
    addfile(files, "", "");
    for (;;) {
        switch (c= *pat++) {

        case EOS:
        case SEP:
            *end= EOS;
            if (meta) {
                if (basestart == mid)
                    segment(files, "", basestart);
                else if (basestart == mid+1) {
                    static char sepstr[]= {SEP, EOS};
                    segment(files, sepstr, basestart);
                }
                else {
                    basestart[-1]= EOS;
                    segment(files, mid, basestart);
                }
                if (files->len == 0)
                    return;
                end= basestart= mid;
                meta= NO;
            }
            else if (c == EOS)
                findfiles(files, mid);
            if (c == EOS)
                return;
            *end++= c;
            basestart= end;
            break;

        case M_ALL:
        case M_ONE:
        case M_SET:
            meta= YES;
            /* Fall through */
        default:
            *end++ = c;

        }
    }
}

/*
 * The main 'glob' routine: does $ and ~ substitution and \ quoting,
 * and then calls 'glob1' to do the pattern matching.
 * Returns 0 if file not found, number of matches otherwise.
 * The matches found are appended to the buffer, separated by
 * EOS characters.  If no matches were found, the pattern is copied
 * to the buffer after $ and ~ substitution and \ quoting.
 */

int
glob(pat, buf, size)
    char *pat;
    char *buf;
    unsigned int size; /* sizeof buf */
{
    char *p, *q;
    char c;
    struct flist files;
    char *start= buf;
    char *end= buf+size;

#if defined(DOS) || defined(OS2)
    fixstr(pat);
#endif
    c= *pat;
    if (c == '~') {
        p= ++pat;
        while (*p != EOS && *p != SEP)
            ++p;
        if (p == pat) {
            q= getenv("HOME");
            if (q == NULL)
                --pat;
            else
                buf= addstr(buf, q, end);
        }
        else {
            q= gethome(makestr(pat, p));
            if (q == NULL)
                --pat;
            else {
                buf= addstr(buf, q, end);
                pat= p;
            }
        }
    }
    else if (c == '$') {
        p= ++pat;
        while (isalnum(*p) || *p == '_')
            ++p;
        q= getenv(makestr(pat, p));
        if (q != NULL) {
            buf= addstr(buf, q, end);
        }
        pat= p;
    }
    else if (c == QUOTE && (pat[1] == '$' || pat[1] == '~'))
        ++pat;
    
    while (buf < end && (c= *pat++) != EOS) {
        switch (c) {
        
        case QUOTE:
            if ((c= *pat++) != EOS && index("\\*?[", c) != NULL)
                *buf++ = c;
            else {
                *buf++ = QUOTE;
                --pat;
            }
            break;
        
        case '*':
            *buf++ = M_ALL;
            break;
        
        case '?':
            *buf++ = M_ONE;
            break;
        
        case '[':
            if (*pat == EOS || index(pat+1, ']') == NULL) {
                *buf++ = c;
                break;
            }
            *buf++ = M_SET;
            c= *pat++;
            do {
                *buf++ = c;
                if (*pat == '-' && (c= pat[1]) != ']') {
                    *buf++ = M_RNG;
                    *buf++= c;
                    pat += 2;
                }
            } while ((c= *pat++) != ']');
            *buf++ = M_END;
            break;
        
        default:
            *buf++ = c;
            break;

        }
    }
    *buf= EOS;
    
    glob1(start, &files);
    
    if (files.len == 0) {
        /* Change meta characters back to printing characters */
        for (buf= start; *buf != EOS; ++buf) {
            if (*buf & 0200)
                *buf &= ~0200;
        }
        return 0; /* No match */
    }
    else {
        int i, len;
        
        qsort((char*)files.item, files.len, sizeof(char*), compare);
        buf= start;
        *buf= EOS;
        for (i= 0; i < files.len; ++i) {
            len= strlen(files.item[i]);
            if (len+1 > size)
                break;
            strcpy(buf, files.item[i]);
            buf += len+1;
            size -= len+1;
        }
        discard(&files);
        return i;
    }
}

#endif /* (not JSTEVIE) or (GLOB is defined) */
