/* $Header: /nw/tony/src/stevie/src/RCS/help.c,v 1.9 89/08/06 09:50:09 tony Exp $
 *
 * Routine to display a command summary.
 */

#include "stevie.h"

extern	uchar	*Version;

#ifdef	HELP

static	char *helpstr[] = {

/***********************************************************************
 * First Screen:   Positioning within file, Adjusting the Screen
 ***********************************************************************/

   "Positioning within file",
   "=======================",
   "^F		Forward screenfull		    Developed by:",
   "^B		Backward screenfull			Tony Andrews",
   "^D		scroll down half screen		    Based on a program by:",
   "^U		scroll up half screen			Tim Thompson",
   "G		Goto line (end default)		    Kanji handling by:",
   "]]		next function				Junn Ohta",
   "[[		previous function",
   "/re		next occurence of regular expression 're'",
   "?re		prior occurence of regular expression 're'",
   "n		repeat last / or ?",
   "N		reverse last / or ?",
   "%		find matching (, ), {, }, [, or ]",
   "",
   "Adjusting the screen",
   "====================",
   "^L		Redraw the screen",
   "^E		scroll window down 1 line",
   "^Y		scroll window up 1 line",
   "z<RETURN>	redraw, current line at top",
   "z-		... at bottom",
   "z.		... at center",
   NULL,

/***********************************************************************
 * Second Screen:   Character positioning
 ***********************************************************************/

   "Character Positioning",
   "=====================",
   "^		first non-white",
   "0		beginning of line",
   "$		end of line",
   "h		backward",
   "l		forward",
   "^H		same as h",
   "space	same as l",
   "fx		find 'x' forward",
   "Fx		find 'x' backward",
   "tx		upto 'x' forward",
   "Tx		upto 'x' backward",
   ";		Repeat last f, F, t, or T",
   ",		inverse of ;",
   "|		to specified column",
   "%		find matching (, ), {, }, [, or ]",
   NULL,

/***********************************************************************
 * Third Screen:   Line Positioning, Marking and Returning
 ***********************************************************************/

   "Line Positioning",
   "================",
   "H		home window line",
   "L		last window line",
   "M		middle window line",
   "+		next line, at first non-white",
   "-		previous line, at first non-white",
   "CR		return, same as +",
   "j		next line, same column",
   "k		previous line, same column",
   "",
   "Marking and Returning",
   "=====================",
#ifdef NOBACKQ
   "`` or \\\\	previous context",
#else
   "``		previous context",
#endif
   "''		... at first non-white in line",
   "mx		mark position with letter 'x'",
#ifdef NOBACKQ
   "`x or \\x	to mark 'x'",
#else
   "`x		to mark 'x'",
#endif
   "'x		... at first non-white in line",
   "",
   "Undo  &  Redo",
   "=============",
   "u		undo last change",
   "U		restore current line",
   ".		repeat last change",
   NULL,

/***********************************************************************
 * Fourth Screen:   Insert & Replace, 
 ***********************************************************************/

   "Insert and Replace",
   "==================",
   "a		append after cursor",
   "i		insert before cursor",
   "A		append at end of line",
   "I		insert before first non-blank",
   "o		open line below",
   "O		open line above",
   "rx		replace single char with 'x'",
   "R		replace characters",
   "",
   "Words, sentences, paragraphs",
   "============================",
   "w		word forward",
   "b		back word",
   "e		end of word",
   ")		to next sentence (not yet)",
   "}		to next paragraph (not yet)",
   "(		back sentence (not yet)",
   "{		back paragraph (not yet)",
   "W		blank delimited word",
   "B		back W",
   "E		to end of W",
   NULL,

/***********************************************************************
 * Fifth Screen:   Misc. operations, 
 ***********************************************************************/

   "Miscellaneous Commands",
   "======================",
   ":w name	write file 'name'",
   ":w		write back changes",
   ":wq		write and quit",
   ":x		write if modified, and quit",
   "ZZ		write if modified, and quit",
   ":q		quit",
   ":q!		quit, discard changes",
   ":e name	edit file 'name'",
   ":e!		edit, discard changes",
   ":e #	edit alternate file",
   ":n		edit next file in arglist",
   ":n files	restart with new arglist",
   ":N		edit prior file in arglist",
   ":rew	rewind arglist",
   ":f		show current file and lines",
   "^G		show current file and lines",
   ":f file	change current file name",
   ":ta tag	to tag file entry 'tag'",
   "^]		:ta, current word is tag",
   ":pop	pop tag stack entry",
   "^T		same as ':pop'",
   NULL,

/***********************************************************************
 * Sixth Screen:   Misc. Operations, (Continued)
 ***********************************************************************/

   "Miscellaneous Commands",
   "======================",
   ":g/pat/p	global search and print",
   ":g/pat/d	global search and delete",
   ":g/pat/s///	global search and substitute",
   ":v/pat/cmd	global search and 'cmd' on non-matching line",
   ":s/p1/p2/	text substitution (trailing 'g' optional)",
   ":d		delete lines",
   ":y		yank lines",
   ":<		shift lines left",
   ":>		shift lines right",
   ":mo d	move lines to line d",
   ":co d	copy lines to line d",
   ":t d	same as ':co d'",
   ":ab word s	map word to string (insert mode)",
   ":map key s	map key to string (command mode)",
   ":map! key s	map key to string (insert mode)",
   ":unab word	undefine abbreviated word",
   ":unmap key	undefine mapped key",
   ":unmap! key	undefine map!ed key",
   ":sh		run an interactive shell",
   ":!cmd	execute a shell command",
   ":r!cmd	read output of command",
   ":so name	source commands from file name",
   NULL,

/***********************************************************************
 * Seventh Screen:   Operators, Misc. operations, Yank & Put
 ***********************************************************************/

   "Operators (double to affect lines)",
   "==================================",
   "d		delete",
   "c		change",
   "<		left shift",
   ">		right shift",
   "y		yank to buffer",
   "!		filter lines",
   "",
   "Miscellaneous operations",
   "========================",
   "C		change rest of line",
   "D		delete rest of line",
   "s		substitute chars",
   "S		substitute lines (not yet)",
   "J		join lines",
   "x		delete characters",
   "X		... before cursor",
   "",
   "Yank and Put",
   "============",
   "p		put back text",
   "P		put before",
   "Y		yank lines",
   NULL,

   NULL
};


bool_t
help()
{
	int	row, col;
	char	**p, *s;
	bool_t	first;

	first = TRUE;
	p = helpstr;
	for (;;) {
		toutstr(T_ED, Rows);
		if (first) {
			windgoto(0, 52);
			outstr(Version);
			first = FALSE;
		}
		row = 0;
		while ((s = *p++) != NULL) {
			col = 4;
			windgoto(row, col);
			while (*s) {
				if (*s == TAB) {
					do {
						outchar(' ');
					} while (++col & 7);
					s++;
					continue;
				}
				outchar(*s++);
				col++;
			}
			row++;
		}
		if (*p == NULL) {
			break;
		}
		windgoto(Rows-2, 52);
		outstr("<Press space to continue>");
		windgoto(Rows-1, 52);
		outstr("<Any other key will quit>");
		if (vgetc() != ' ')
			return TRUE;
	}
	windgoto(Rows-1, 52);
	outstr("<Press any key>");
	vgetc();
	return TRUE;
}

#else

bool_t
help()
{
	msg("Sorry, help not configured");
	return FALSE;
}
#endif
