/* $Header: /nw/tony/src/stevie/src/RCS/minix.c,v 1.5 89/07/11 21:24:18 tony Exp $
 *
 * System-dependent routines for Minix-ST
 */

#include "stevie.h"
#include <sgtty.h>
#include <signal.h>
#include <stat.h>
#include <errno.h>

#ifndef MAXNAMLEN
#define	MAXNAMLEN	14
#endif

uchar	*Initfile = ".stevierc";

/*
 * inchar() - get a character from the keyboard
 */
int
inchar()
{
	uchar	c;

	flushbuf();		/* flush any pending output */

	while (read(0, &c, 1) != 1)
		;

	got_int = FALSE;
	return c;
}

static	uchar	outbuf[OBSIZE];
static	int	bpos = 0;

void
flushbuf()
{
	uchar * p = outbuf;
	int writed;

	while (bpos) {
		writed = write(2, p, bpos);
		if (writed < 0)
			continue;
		bpos -= writed;
		p += writed;
	}
}

/*
 * Macro to output a character. Used within this file for speed.
 */
#define	outone(c)	outbuf[bpos++] = c; if (bpos >= OBSIZE) flushbuf()

/*
 * Function version for use outside this file.
 */
void
outchar(c)
register uchar	c;
{
	outbuf[bpos++] = c;
	if (bpos >= OBSIZE)
		flushbuf();
}

void
outstr(s)
register uchar	*s;
{
	while (*s) {
		outone(*s++);
	}
}

void
beep()
{
	outone('\007');
}

void
delay()
{
	/* not implemented */
}

static	struct	sgttyb	ostate;

/*
 * Go into cbreak mode
 */
void
set_tty()
{
	struct	sgttyb	nstate;

	 ioctl(0, TIOCGETP, &ostate);
	 nstate = ostate;
	 nstate.sg_flags &= ~(XTABS|ECHO);
	 nstate.sg_flags |= CBREAK;
	 ioctl(0, TIOCSETP, &nstate);
}

/*
 * Restore original terminal modes
 */
void
reset_tty()
{
	ioctl(0, TIOCSETP, &ostate);
}

void
sig()
{
	signal(SIGINT, sig);
	signal(SIGQUIT, sig);

	got_int = TRUE;
}

void
tstpon()
{
	/* nothing to do here */
}

void
tstpoff()
{
	/* nothing to do here */
}

void
windinit()
{
#ifdef	TERMCAP
	if (t_init() != 1) {
		fprintf(stderr, "unknown terminal type\n");
		exit(1);
	}
#else
	Columns = 80;
	P(P_LI) = Rows = 25;
#endif
	/*
	 * The code here makes sure that there isn't a window during which
	 * we could get interrupted and exit with the tty in a weird state.
	 */
	signal(SIGINT, sig);
	signal(SIGQUIT, sig);

	set_tty();

	if (got_int)
		windexit(0);

	if (T_TI) toutstr(T_TI, 1);
}

void
windexit(r)
int r;
{
	if (T_TE) toutstr(T_TE, 1);
	flushbuf();
	reset_tty();
	exit(r);
}

void
windgoto(r, c)
register int	r, c;
{
#ifdef	TERMCAP
	char	*tgoto();

	toutstr(tgoto(T_CM, c, r), 1);
#else
	r++;
	c++;

	/*
	 * Check for overflow once, to save time.
	 */
	if (bpos + 8 >= OBSIZE)
		flushbuf();

	outbuf[bpos++] = '\033';
	outbuf[bpos++] = '[';
	if (r >= 10) {
		outbuf[bpos++] = r / 10 + '0';
		outbuf[bpos++] = r % 10 + '0';
	} else
		outbuf[bpos++] = r + '0';
	outbuf[bpos++] = ';';
	if (c >= 10) {
		outbuf[bpos++] = c / 10 + '0';
		outbuf[bpos++] = c % 10 + '0';
	} else
		outbuf[bpos++] = c + '0';
	outbuf[bpos++] = 'H';
#endif
}

char *
strchr(s, c)
char	*s;
char	c;
{
	char *index();

	return index(s, c);
}

FILE *
fopenb(fname, mode)
uchar	*fname;
uchar	*mode;
{
	return fopen(fname, mode);
}

int
filestat(fname)
uchar	*fname;
{
	int	fd, ret;
	struct stat stbuf;
	extern int errno;

	fd = open(fname, 0);
	if (fd == -1) {
		if (errno == ENOENT)
			ret = 0;
		else
			ret = FS_EXIST;
	} else {
		ret = FS_EXIST | FS_READ;
		fstat(fd, &stbuf);
		if (stbuf.st_mode & S_IFDIR)
			ret |= FS_DIR;
		close(fd);
		if (access(fname, 2) == 0)
			ret |= FS_WRITE;
	}
	return ret;
}

/*
 * remove(file) - remove a file
 */
int
remove(file)
uchar *file;
{
	return unlink(file);
}

/*
 * Rename(of, nf) - rename existing file 'of' to 'nf'
 */
int
Rename(of, nf)
uchar	*of, *nf;
{
	unlink(nf);
	if (link(of, nf) != -1 && unlink(of) != -1)
		return 0;
	return -1;
}

/*
 * backupname(fname, ext) - generate filename for backup
 */
uchar *
backupname(fname, ext)
uchar	*fname, *ext;
{
	register uchar	*s, *top;
	int		elen;
	static uchar	bname[MAXPLEN];

	elen = strlen(ext);
	strcpy(bname, fname);
	for (s = bname; *s; s++)
		/* void */;
	while (s > bname && *s != '/')
		s--;
	top = ++s;
	while (*s && (s - top) + elen < MAXNAMLEN)
		s++;
	*s = NUL;
	strcat(bname, ext);
	return bname;
}

uchar *
fixname(s)
uchar	*s;
{
	return s;
}

uchar *
mkpath(dir, file)
uchar	*dir, *file;
{
	int	n;
	static uchar	buf[MAXPLEN];

	n = 0;
	if (dir) {
		strcpy(buf, dir);
		n = strlen(buf);
		if (buf[n-1] != '/')
			buf[n++] = '/';
	}
	strcpy(buf+n, file);
	return buf;
}

/*
 * doshell() - run a command or an interactive shell
 */
void
doshell(cmd)
uchar	*cmd;
{
	char	*cp, *getenv();
	uchar	cline[MAXCLIN];

	outstr("\r\n");
	flushbuf();

	if (cmd == NULL) {
		if ((cmd = getenv("SHELL")) == NULL)
			cmd = "/bin/sh";
		sprintf(cline, "%s -i", cmd);
		cmd = cline;
	}

	reset_tty();
	system(cmd);
	set_tty();

	wait_return();
}

void
cmdread(cmd)
uchar	*cmd;
{
	int fd, old, err;
	char tmpfile[MAXPLEN];

	if (cmd[0] == NUL) {
		emsg("Exec failed");
		return;
	}
	sprintf(tmpfile, "/tmp/vi%d", getpid());
	if ((fd = creat(tmpfile, 0600)) < 0) {
		emsg("Can't create temporary file");
		return;
	}
	old = dup(1);
	dup2(fd, 1);
	close(fd);
	outstr("\r");
	flushbuf();
	reset_tty();
	err = system(cmd);
	set_tty();
	close(1);
	dup2(old, 1);
	if (err < 0)
		emsg("Exec failed");
	else {
		if (readfile(tmpfile, Curschar, 1))
			emsg("Can't open temporary file");
		else {
			updatescreen();
			CHANGED;
		}
	}
	remove(tmpfile);
}
