/* $Header: /nw/tony/src/stevie/src/RCS/param.c,v 1.10 89/08/02 10:59:10 tony Exp $
 *
 * Code to handle user-settable parameters. This is all pretty much table-
 * driven. To add a new parameter, put it in the params array, and add a
 * macro for it in param.h. If it's a numeric parameter, add any necessary
 * bounds checks to doset(). String parameters aren't currently supported.
 */

#include "stevie.h"

struct	param	params[] = {
	{ "autoindent",	"ai",		FALSE,	P_BOOL },
	{ "autowrite",	"aw",		FALSE,	P_BOOL },
	{ "backup",	"bk",		FALSE,	P_BOOL },
	{ "errorbells",	"eb",		FALSE,	P_BOOL },
#ifdef FEPCTRL
	{ "fepcontrol",	"fc",		FALSE,	P_BOOL },
#endif
	{ "ignorecase",	"ic",		FALSE,	P_BOOL },
	{ "lines",	"lines",	25,	P_NUM },
	{ "list",	"list",		FALSE,	P_BOOL },
	{ "modelines",	"ml",		FALSE,	P_BOOL },
	{ "number",	"nu",		FALSE,	P_BOOL },
	{ "readonly",	"ro",		FALSE,	P_BOOL },
	{ "redraw",	"redraw",	TRUE,	P_BOOL },
	{ "report",	"report",	5,	P_NUM },
#ifdef CPMSTYLE
	{ "return",	"cr",		TRUE,	P_BOOL },
#else
	{ "return",	"cr",		FALSE,	P_BOOL },
#endif
	{ "scroll",	"scroll",	12,	P_NUM },
 	{ "shiftwidth",	"sw",		8,	P_NUM },
	{ "showmatch",	"sm",		FALSE,	P_BOOL },
	{ "showmode",	"mo",		FALSE,	P_BOOL },
	{ "tabstop",	"ts",		8,	P_NUM },
	{ "terse",	"terse",	FALSE,	P_BOOL },
	{ "tildeop",	"to",		FALSE,	P_BOOL },
	{ "vbell",	"vb",		TRUE,	P_BOOL },
	{ "wrapscan",	"ws",		TRUE,	P_BOOL },
	{ "",		"",		0,	0, }		/* end marker */

};

static	void	showparms();

void
doset(arg)
char	*arg;		/* parameter string */
{
	register int	i;
	register char	*s;
	bool_t	did_lines = FALSE;
	bool_t	state = TRUE;		/* new state of boolean parms. */
	int	sep;
#ifdef FEPCTRL
	int	old_fc = P(P_FC);
#endif

	if (arg == NULL) {
		showparms(FALSE);
		return;
	}
	if (strncmp(arg, "all", 3) == 0) {
		showparms(TRUE);
		return;
	}
	if (strncmp(arg, "no", 2) == 0) {
		state = FALSE;
		arg += 2;
	}

	for (i=0; params[i].fullname[0] != NUL ;i++) {
		s = params[i].fullname;
		if (strncmp(arg, s, strlen(s)) == 0)	/* matched full name */
			break;
		s = params[i].shortname;
		if (strncmp(arg, s, strlen(s)) == 0)	/* matched short name */
			break;
	}

	if (params[i].fullname[0] != NUL) {	/* found a match */
		if (params[i].flags & P_NUM) {
			did_lines = (i == P_LI);
			sep = arg[strlen(s)];
			if (sep != '=' && sep != ':' || state == FALSE)
				emsg("Invalid set of numeric parameter");
			else {
				params[i].value = atoi(arg+strlen(s)+1);
				params[i].flags |= P_CHANGED;
			}
		} else /* boolean */ {
			sep = arg[strlen(s)];
			if (sep == '=' || sep == ':')
				emsg("Invalid set of boolean parameter");
			else {
				params[i].value = state;
				params[i].flags |= P_CHANGED;
			}
		}
	} else
		emsg("Unrecognized 'set' option");

#ifdef FEPCTRL
	if (P(P_FC) != old_fc) {
		if (old_fc)
			fep_term();
		else
			fep_init();
	}
#endif

	/*
	 * Update the screen in case we changed something like "tabstop"
	 * or "list" that will change its appearance.
	 */
	updatescreen();

	if (did_lines) {
		Rows = P(P_LI);
		P(P_SS) = Rows / 2;
		screenalloc();		/* allocate new screen buffers */
		screenclear();
		updatescreen();
	}
	/*
	 * Check the bounds for numeric parameters here
	 */
	if (P(P_TS) <= 0 || P(P_TS) > 32) {
		emsg("Invalid tab size specified");
		P(P_TS) = 8;
		return;
	}

	if (P(P_SW) <= 0 || P(P_SW) > 32) {
		emsg("Invalid shift size specified");
		P(P_SW) = 4;
		return;
	}

	if (P(P_SS) <= 0 || P(P_SS) > Rows) {
		emsg("Invalid scroll size specified");
		P(P_SS) = Rows / 2;
		return;
	}

#ifndef	TILDEOP
	if (P(P_TO)) {
		emsg("Tilde-operator not enabled");
		P(P_TO) = FALSE;
		return;
	}
#endif
	/*
	 * Check for another argument, and call doset() recursively, if
	 * found. If any argument results in an error, no further
	 * parameters are processed.
	 */
	while (*arg != ' ' && *arg != '\t') {	/* skip to next white space */
		if (*arg == NUL)
			return;			/* end of parameter list */
		arg++;
	}
	while (*arg == ' ' || *arg == '\t')	/* skip to next non-white */
		arg++;

	if (*arg)
		doset(arg);	/* recurse on next parameter */
}

static uchar *
oneparm(p, len)
struct param	*p;
int	len;
{
	int	i;
	static	uchar	buf[MAXMLEN];

	if (p->flags & P_BOOL)
		sprintf(buf, "%s%s", (p->value? "": "no"), p->fullname);
	else
		sprintf(buf, "%s=%d", p->fullname, p->value);
	i = strlen(buf);
	while (i < len)
		buf[i++] = ' ';
	buf[i] = NUL;
	return buf;
}

static	void
showparms(all)
bool_t	all;	/* show ALL parameters */
{
	register struct	param	*p;
	int	i, j, n;
	char	buf[MAXMLEN];

	if (!all) {
		*buf = NUL;
		for (p = params; *p->fullname; p++) {
			if (!(p->flags & P_CHANGED))
				continue;
			if (*buf)
				strcat(buf, " ");
			strcat(buf, oneparm(p, 1));
		}
		msg(buf);
		return;
	}
	gotocmd(FALSE, 0);
	outstr("\r\n");
	for (n = 0, p = params; *p->fullname; p++, n++)
		;
	j = (n + 2) / 3;
	for (i = 0; i < j; i++) {
		*buf = NUL;
		if (i+j+j < n) {
			strcat(buf, oneparm(params+i, 24));
			strcat(buf, oneparm(params+i+j, 24));
			strcat(buf, oneparm(params+i+j+j, 1));
		} else {
			strcat(buf, oneparm(params+i, 24));
			strcat(buf, oneparm(params+i+j, 1));
		}
		strcat(buf, "\r\n");
		outstr(buf);
	}
	wait_return();
}
