/*
 * pc98_mou.c - NEC PC-9801 dependent mouse handling routines
 *
 *	Written by: Kazuhisa Hayashi, February 1990
 *	Arranged by: Junn Ohta, May 1990
 */
#include "env.h"
#include "stevie.h"

#ifdef	PC98_MOUSE

#include <stdio.h>
#include <dos.h>
#include "pc98_mou.h"
#include "pc98_cbm.h"

static	int	mouse_install;

mouse_init()
{
	char	*mouse_map = bitmap;

	mouse_install = MouseInit();
	if (mouse_install) {
		MouseRange(0, 639, 0, 399-16);
		MouseSetPos(635, 8);	/* center of top line, last column */
		MouseCsrSet(MOUSE_SPOT_H, MOUSE_SPOT_V,
				FP_SEG(mouse_map), FP_OFF(mouse_map));
		MouseCsrPlane(3);
		MouseCsrOn();
	}
}

mouse_exit()
{
	if (mouse_install)
		MouseCsrOff();
}

unsigned short
jistojms(c)
register unsigned short c;
{
	register unsigned short c1;

	c1 = c >> 8;
	c &= 0xff;
	if (c1 < 0x21 || c1 > 0x7f || c < 0x21 || c > 0x7f)
		return 0;
	c += (c1 & 0x01)? (c > 0x5f)? 0x20: 0x1f: 0x7e;
	c1 = (c1 - 0x21) >> 1;
	c1 += (c1 > 0x1e)? 0xc1: 0x81;
	return (c1 << 8) | c;
}

#define	MK_FAR(s,o) (unsigned char far *)((unsigned long)(s)<<16|(unsigned)(o))

static	unsigned short far *vram  = (unsigned short far *)MK_FAR(0xa000, 0);
static	unsigned short far *vattr = (unsigned short far *)MK_FAR(0xa200, 0);

AttrSet(x, y, attr)
int x, y, attr;
{
	if (x < 0 || y < 0)
		return;
	*(unsigned char far *)&vattr[y * 80 + x] = attr;
}

AttrGet(x, y, attrp)
int	x, y;
unsigned char	*attrp;
{
	if (x < 0 || y < 0)
		return;
	*attrp = *(unsigned char far *)&vattr[y * 80 + x];
}

PosToCol(xpos, ypos, xcol, ycol, start)
int	xpos, ypos, *xcol, *ycol, start;
{
	unsigned short	c;

	*xcol = xpos >> 3;
	*ycol = ypos >> 4;
	c = vram[(*ycol) * 80 + (*xcol)];
	if ((c & 0xff00) == 0)
		return;
	if (c & 0x8000) {
		if (start && *xcol > 0)
			(*xcol)--;
	} else {
		if (!start && *xcol < 79)
			(*xcol)++;
	}
}

VramToString(x0, y0, x1, y1, yankbuf, yanklenp)
int	x0, y0, x1, y1;
char	*yankbuf;
short	*yanklenp;
{
	register int	i;
	register int	last;	/* last non-blank column on the line */
	int		next;	/* first column of the next line */
	int		x, y;	/* current (start) and end column */
	unsigned short	c;

	x = y0 * 80 + x0;
	y = y1 * 80 + x1;
	if (x > y) {
		i = x;
		x = y;
		y = i;
	}
	next = x - (x % 80) + 80;
	i = 0;
	for (;;) {
		last = next - 1;
		while (last >= x && last != y && vram[last] == 0x0020)
			last--;
		while (x <= last && x <= y) {
			c = vram[x++];
			if (c & 0xff00) {
				if (i + 1 >= YANK_MAX)
					goto overflow;
				c &= 0x7f7f;
				c += 0x0020;
				c = jistojms(c << 8 | c >> 8);
				yankbuf[i++] = c >> 8;
				yankbuf[i++] = c;
				x++;
			} else {
				if (i >= YANK_MAX)
					goto overflow;
				yankbuf[i++] = c;
			}
		}
		if (x > y)
			break;
		if (last < next - 1) {
			if (i >= YANK_MAX)
				goto overflow;
			yankbuf[i++] = '\n';
		}
		x = next;
		next += 80;
	}
overflow:
	yankbuf[i] = 0;
	*yanklenp = i;
}

static	union	REGS	regs;
static	struct	SREGS	sregs;

MouseInit()
{
	regs.x.ax = 0;
	int86(0x33, &regs, &regs);
	return regs.x.ax;
}

MouseRange(xs, xe, ys, ye)
int	xs, xe, ys, ye;
{
	if (0 <= xs && xs <= xe && xe < 640) {
		regs.x.ax = 16;
		regs.x.cx = xs;
		regs.x.dx = xe;
		int86(0x33, &regs, &regs);
	}
	if (0 <= ys && ys <= ye && ye < 400) {
		regs.x.ax = 17;
		regs.x.cx = ys;
		regs.x.dx = ye;
		int86(0x33, &regs, &regs);
	}
}

MouseMove(xret, yret)
short	*xret, *yret;
{
	regs.x.ax = 3;
	int86(0x33, &regs, &regs);
	*xret = regs.x.cx;
	*yret = regs.x.dx;
	regs.h.al = regs.h.bl;
	return regs.x.ax;
}

MouseSetPos(x, y)
short	x, y;
{
	if (0 <= x && x < 640 && 0 <= y && y < 400) {
		regs.x.ax = 4;
		regs.x.cx = x;
		regs.x.dx = y;
		int86(0x33, &regs, &regs);
	}
}

MouseDownL(xret, yret)
short	*xret, *yret;
{
	regs.x.ax = 5;
	int86(0x33, &regs, &regs);
	if (regs.x.ax) {
		*xret = regs.x.cx;
		*yret = regs.x.dx;
	}
	return regs.x.ax;
}

MouseReleaseL(xret, yret)
short	*xret, *yret;
{
	regs.x.ax = 6;
	int86(0x33, &regs, &regs);
	if (regs.x.ax)
		return 0;
	*xret = regs.x.cx;
	*yret = regs.x.dx;
	return 1;
}

MouseDownR(xret, yret)
short	*xret, *yret;
{
	regs.x.ax = 7;
	int86(0x33, &regs, &regs);
	if (regs.x.ax) {
		*xret = regs.x.cx;
		*yret = regs.x.dx;
	}
	return regs.x.ax;
}

MouseReleaseR(xret, yret)
short	*xret, *yret;
{
	do {
		regs.x.ax = 8;
		int86(0x33, &regs, &regs);
	} while (regs.x.ax);
	*xret = regs.x.cx;
	*yret = regs.x.dx;
}

MouseCsrOn()
{
	regs.x.ax = 1;
	int86(0x33, &regs, &regs);
}

MouseCsrOff()
{
	regs.x.ax = 2;
	int86(0x33, &regs, &regs);
}

MouseCsrSet(pot_h, pot_v, seg, off)
int	pot_h, pot_v, seg, off;
{
	regs.x.ax = 9;
	regs.x.bx = pot_h;
	regs.x.cx = pot_v;
	regs.x.dx = off;
	sregs.es  = seg;
	int86x(0x33, &regs, &regs, &sregs);
}

MouseCsrPlane(plane)
int	plane;
{
	regs.x.ax = 18;
	regs.x.bx = plane & 03;
	int86(0x33, &regs, &regs);
}

#define	getchnowait()	(bdos(6, 0xff, 0) & 0xff)

NextEvent(eventp)
union	event_t	*eventp;
{
	unsigned char	c;
	int	x, y;

	for (;;) {
		if ((c = getchnowait()) != 0) {
			eventp->key.type = KEYINPUT;
			eventp->key.code = c;
			break;
		}
		if (mouse_install && MouseDownL(&x, &y)) {
			eventp->mouse.type = MOUSE_DOWN_L;
			eventp->mouse.x = x;
			eventp->mouse.y = y;
			break;
		}
		if (mouse_install && MouseDownR(&x, &y)) {
			eventp->mouse.type = MOUSE_DOWN_R;
			eventp->mouse.x = x;
			eventp->mouse.y = y;
			break;
		}
	}
}

#endif	/* PC98_MOUSE */
