/* $Header: /nw/tony/src/stevie/src/RCS/search.c,v 1.16 89/08/06 09:50:51 tony Exp $
 *
 * This file contains various searching-related routines. These fall into
 * three groups: string searches (for /, ?, n, and N), character searches
 * within a single line (for f, F, t, T, etc), and "other" kinds of searches
 * like the '%' command, and 'word' searches.
 */

#include "stevie.h"
#include "regexp.h"	/* Henry Spencer's (modified) reg. exp. routines */

static	char	*noprev = "No previous regular expression";
static	char	*notfound = "Pattern not found";
static	char	*invalid = "Invalid search string";
static	char	*invcom = "Invalid command character";
static	char	*nullpat = "NULL pattern specified";
static	char	*nomatch = "No match";
static	char	*intrpt = "Interrupt";
static	char	*ovrflw = "Line overflow in substitute";

/*
 * String searches
 *
 * The actual searches are done using Henry Spencer's regular expression
 * library.
 */

#define	BEGWORD	"([^a-zA-Z0-9_]|^)"	/* replaces "\<" in search strings */
#define	ENDWORD	"([^a-zA-Z0-9_]|$)"	/* likewise replaces "\>" */

#ifdef KANJI
#define	BEGCHAR(c) \
	(((c) & 0x80) || isalnum(c) || ((c) == '_'))
#else
#define	BEGCHAR(c)	(isalnum(c) || ((c) == '_'))
#endif

bool_t	reg_ic;		/* ignore case in searches */

bool_t	begword;	/* does the search include a 'begin word' match */

static	bool_t	nomem = FALSE;	/* can't allocate memory */

/*
 * foldcase(s) - fold letters into uppercase in the string
 */
static uchar *
foldcase(s)
register uchar	*s;
{
	static uchar	ns[MAXLINE];
	register uchar	*p;

	p = ns;
	while (*s) {
#ifdef KANJI
		if (isKanji(*s)) {
			*p++ = *s++;
			*p++ = *s++;
		} else if (*s & 0x80)
			*p++ = *s++;
		else
#endif
		if (islower(*s))
			*p++ = toupper(*s++);
		else
			*p++ = *s++;
	}
	*p = NUL;
	return ns;
}

/*
 * mapstring(s) - map special backslash sequences
 */
static uchar *
mapstring(s)
register uchar	*s;
{
	static	uchar	ns[MAXCLIN];
	register uchar	*p;

	begword = FALSE;

	for (p = ns; *s ;s++) {
#ifdef KANJI
		if (isKanji(*s)) {	/* first byte of kanji */
			*p++ = *s++;
			*p++ = *s;
			continue;
		}
#endif
		if (*s == '(' || *s == ')') {
			*p++ = '\\';
			*p++ = *s;
			continue;
		}
		if (*s != '\\') {	/* not an escape */
			*p++ = *s;
			continue;
		}
		switch (*++s) {
		case '/':
		case '(':
		case ')':
			*p++ = *s;
			break;

		case '<':
			strcpy(p, BEGWORD);
			p += strlen(BEGWORD);
			begword = TRUE;
			break;

		case '>':
			strcpy(p, ENDWORD);
			p += strlen(ENDWORD);
			break;

		default:
			*p++ = '\\';
			*p++ = *s;
			break;
		}
	}
	*p = NUL;

	return ns;
}

static uchar *laststr = NULL;
static int lastsdir;

static LPTR *
ssearch(dir,str)
int	dir;	/* FORWARD or BACKWARD */
uchar	*str;
{
	LPTR	*bcksearch(), *fwdsearch();
	LPTR	*pos;
	uchar	*old_ls = laststr;
	uchar	*ls;
	int	col;

	reg_ic = P(P_IC);

	laststr = strsave(str);
	if (laststr == NULL) {
		nomem = TRUE;
		return (LPTR *)NULL;
	}
	lastsdir = dir;

	ls = reg_ic? foldcase(laststr): laststr;

	if (old_ls != NULL)
		free(old_ls);

	if (dir == BACKWARD) {
		col = 1;
		smsg("?%s", pretty(laststr, &col));
		windgoto(Rows-1, 0);
		flushbuf();
		pos = bcksearch(mapstring(ls));
	} else {
		col = 1;
		smsg("/%s", pretty(laststr, &col));
		windgoto(Rows-1, 0);
		flushbuf();
		pos = fwdsearch(mapstring(ls));
	}

	/*
	 * This is kind of a kludge, but its needed to make
	 * 'beginning of word' searches land on the right place.
	 */
	if (pos != NULL && begword) {
		if (pos->index != 0 || !BEGCHAR(pos->linep->s[0]))
			pos->index += 1;
	}
	return pos;
}

bool_t
dosearch(dir,str)
int	dir;
uchar	*str;
{
	LPTR	*p;

	if (str == NULL) {
		if (laststr == NULL) {
			emsg(noprev);
			return FALSE;
		}
		str = laststr;
	}

	got_int = FALSE;

	if ((p = ssearch(dir,str)) == NULL) {
		if (got_int)
			emsg(intrpt);
		else if (!nomem)
			emsg(notfound);

		got_int = FALSE;
		return FALSE;
	} else {
		cursupdate();
		/*
		 * if we're backing up, we make sure the line we're on
		 * is on the screen.
		 */
		setpcmark();
		*Curschar = *p;
		set_want_col = TRUE;
		cursupdate();

		return TRUE;
	}
}

LPTR *
secretsearch(dir,str)
int	dir;
uchar	*str;
{
	LPTR	*p;

	if (str == NULL) {
		if (laststr == NULL) {
			emsg(noprev);
			return FALSE;
		}
		str = laststr;
	}

	got_int = FALSE;

	if ((p = ssearch(dir,str)) == NULL) {
		if (got_int)
			emsg(intrpt);
		else if (!nomem)
			emsg(notfound);

		got_int = FALSE;
	} 
	return p;
}

#define	OTHERDIR(x)	(((x) == FORWARD) ? BACKWARD : FORWARD)

bool_t
repsearch(flag)
int	flag;
{
	int	dir = lastsdir;
	bool_t	found;

	if ( laststr == NULL ) {
		emsg(noprev);
		return FALSE;
	}

	found = dosearch(flag ? OTHERDIR(lastsdir) : lastsdir, laststr);

	/*
	 * We have to save and restore 'lastsdir' because it gets munged
	 * by ssearch() and winds up saving the wrong direction from here
	 * if 'flag' is true.
	 */
	lastsdir = dir;

	return found;
}

/*
 * regerror - called by regexp routines when errors are detected.
 */
void
regerror(s)
uchar	*s;
{
	emsg(s);
}

#ifdef KANJI
static Char *
normalize(s)
register uchar	*s;
{
	register Char	*u;
	static Char	ubuf[MAXLINE];

	u = ubuf;
	while (*s) {
		if (isKanji(*s)) {
			*u = *s++ << 8;
			*u++ |= *s++;
		} else
			*u++ = *s++;
	}
	*u = 0;
	return ubuf;
}

static Char *
usave(u)
register Char	*u;
{
	register Char	*p, *top;

	for (p = u; *p; p++)
		;
	top = p = (Char *)alloc((unsigned)(p - u + 1) * sizeof(Char));
	if (p == NULL)
		return NULL;
	while (*p++ = *u++)
		;
	return top;
}

static int
ulen(from, to)
register Char	*from, *to;
{
	register int	len;

	len = 0;
	while (from < to) {
		if (*from++ & 0xff00)
			len++;
		len++;
	}
	return len;
}

static uchar *
unnormalize(u)
register Char	*u;
{
	register uchar	*s;
	static uchar	sbuf[MAXLINE];

	s = sbuf;
	while (*u) {
		if (*u & 0xff00)
			*s++ = (uchar)(*u >> 8);
		*s++ = (uchar)*u++;
	}
	*s = 0;
	return sbuf;
}
#endif

static LPTR *
fwdsearch(str)
register uchar	*str;
{
	static LPTR	infile;
	register LPTR	*p;
	regexp	*prog;
	register uchar	*linestr;
	register int	i;
#ifdef KANJI
	register Char	*u;
#else
	register uchar	*s;
#endif

#ifdef KANJI
	if ((prog = regcomp(normalize(str))) == NULL)
#else
	if ((prog = regcomp(str)) == NULL)
#endif
	{
		emsg(invalid);
		return NULL;
	}

	p = Curschar;
	i = p->index;
	if (p->linep->s[i] == NUL) {	/* at empty line */
		p = nextline(p);	/* start at next line */
		i = 0;
	} else {
#ifdef KANJI
		i += isKanji(p->linep->s[i])? 2: 1;
#else
		i++;
#endif
	}

	while (p != NULL) {
		linestr = p->linep->s;
		if (reg_ic)
			linestr = foldcase(linestr);

#ifdef KANJI
		u = normalize(linestr + i);
		if (regexec(prog, u, i == 0)) {		/* got a match */
			infile.linep = p->linep;
			infile.index = i + ulen(u, prog->startp[0]);
			if (infile.index == 0 || *prog->startp[0] != NUL) {
				free((char *)prog);
				return (&infile);
			}
		}
#else
		s = linestr + i;
		if (regexec(prog, s, i == 0)) {		/* got a match */
			infile.linep = p->linep;
			infile.index = (int) (prog->startp[0] - linestr);
			if (infile.index == 0 || *prog->startp[0] != NUL) {
				free((char *)prog);
				return (&infile);
			}
		}
#endif

		if (got_int)
			goto fwdfail;

		p = nextline(p);
		i = 0;
	}

	/*
	 * If wrapscan isn't set, then don't scan from the beginning
	 * of the file. Just return failure here.
	 */
	if (!P(P_WS))
		goto fwdfail;

	/* search from the beginning of the file to Curschar */
	for (p = Filemem; p != NULL ;p = nextline(p)) {
		linestr = p->linep->s;
		if (reg_ic)
			linestr = foldcase(linestr);
#ifdef KANJI
		u = normalize(linestr);
		if (regexec(prog, u, TRUE)) {		/* got a match */
			infile.linep = p->linep;
			infile.index = ulen(u, prog->startp[0]);
			if (infile.index == 0 || *prog->startp[0] != NUL) {
				free((char *)prog);
				return (&infile);
			}
		}
#else
		s = linestr;
		if (regexec(prog, s, TRUE)) {		/* got a match */
			infile.linep = p->linep;
			infile.index = (int) (prog->startp[0] - s);
			if (infile.index == 0 || *prog->startp[0] != NUL) {
				free((char *)prog);
				return (&infile);
			}
		}
#endif
		if (p->linep == Curschar->linep)
			break;

		if (got_int)
			goto fwdfail;
	}

fwdfail:
	free((char *)prog);
	return NULL;
}

static LPTR *
bcksearch(str)
uchar	*str;
{
	static LPTR	infile;
	register LPTR	*p = &infile;
	register uchar	*linestr;
	register int	i;
#ifdef KANJI
	register Char	*u, *match;
#else
	register uchar	*s, *match;
#endif
	regexp	*prog;

	/* make sure str isn't empty */
	if (str == NULL || *str == NUL)
		return NULL;

#ifdef KANJI
	if ((prog = regcomp(normalize(str))) == NULL)
#else
	if ((prog = regcomp(str)) == NULL)
#endif
	{
		emsg(invalid);
		return NULL;
	}

	*p = *Curschar;

#ifdef KANJI
	if (kdec(p) == -1) {	/* already at start of file? */
		*p = *Fileend;
		p->index = strlen(p->linep->s);
		kdec(p);
	}

	if (begword)		/* so we don't get stuck on one match */
		kdec(p);
#else
	if (dec(p) == -1) {	/* already at start of file? */
		*p = *Fileend;
		p->index = strlen(p->linep->s) - 1;
	}

	if (begword)		/* so we don't get stuck on one match */
		dec(p);
#endif

	i = p->index;

	do {
		linestr = p->linep->s;
		if (reg_ic)
			linestr = foldcase(linestr);
#ifdef KANJI
		u = normalize(linestr);
		if (regexec(prog, u, TRUE)) {	/* match somewhere on line */

			/*
			 * Now, if there are multiple matches on this line,
			 * we have to get the last one. Or the last one
			 * before the cursor, if we're on that line.
			 */
			match = prog->startp[0];

			while (regexec(prog, prog->endp[0], FALSE)) {
				if ((i >= 0) && (ulen(u, prog->startp[0]) > i))
					break;
				if (prog->startp[0] == prog->endp[0]) {
					/* found a size 0 match */
					break;
				}
				match = prog->startp[0];
			}

			if ((i >= 0) && (ulen(u, match) > i)) {
				i = -1;
				continue;
			}

			infile.linep = p->linep;
			infile.index = ulen(u, match);
			free((char *)prog);
			return (&infile);
		}
#else
		s = linestr;
		if (regexec(prog, s, TRUE)) {	/* match somewhere on line */

			/*
			 * Now, if there are multiple matches on this line,
			 * we have to get the last one. Or the last one
			 * before the cursor, if we're on that line.
			 */
			match = prog->startp[0];

			while (regexec(prog, prog->endp[0], FALSE)) {
				if ((i >= 0) && ((prog->startp[0] - s) > i))
					break;
				if (prog->startp[0] == prog->endp[0]) {
					/* found a size 0 match */
					break;
				}
				match = prog->startp[0];
			}

			if ((i >= 0) && ((match - s) > i)) {
				i = -1;
				continue;
			}

			infile.linep = p->linep;
			infile.index = (int) (match - s);
			free((char *)prog);
			return (&infile);
		}
#endif
		i = -1;

		if (got_int)
			goto bckfail;

	} while ((p = prevline(p)) != NULL);

	/*
	 * If wrapscan isn't set, bag the search now
	 */
	if (!P(P_WS))
		goto bckfail;

	/* search backward from the end of the file */
	p = prevline(Fileend);
	do {
		linestr = p->linep->s;
		if (reg_ic)
			linestr = foldcase(linestr);
#ifdef KANJI
		u = normalize(linestr);
		if (regexec(prog, u, TRUE)) {	/* match somewhere on line */

			/*
			 * Now, if there are multiple matches on this line,
			 * we have to get the last one.
			 */
			match = prog->startp[0];

			while (regexec(prog, prog->endp[0], FALSE)) {
				if (prog->startp[0] == prog->endp[0]) {
					/* found a size 0 match */
					break;
				}
				match = prog->startp[0];
			}

			infile.linep = p->linep;
			infile.index = ulen(u, match);
			free((char *)prog);
			return (&infile);
		}
#else
		s = linestr;
		if (regexec(prog, s, TRUE)) {	/* match somewhere on line */

			/*
			 * Now, if there are multiple matches on this line,
			 * we have to get the last one.
			 */
			match = prog->startp[0];

			while (regexec(prog, prog->endp[0], FALSE)) {
				if (prog->startp[0] == prog->endp[0]) {
					/* found a size 0 match */
					break;
				}
				match = prog->startp[0];
			}

			infile.linep = p->linep;
			infile.index = (int) (match - s);
			free((char *)prog);
			return (&infile);
		}
#endif
		if (p->linep == Curschar->linep)
			break;

		if (got_int)
			goto bckfail;

	} while ((p = prevline(p)) != NULL);

bckfail:
	free((char *)prog);
	return NULL;
}

/*
 * dosub(lp, up, cmd)
 *
 * Perform a substitution from line 'lp' to line 'up' using the
 * command pointed to by 'cmd' which should be of the form:
 *
 * /pattern/substitution/g
 *
 * The trailing 'g' is optional and, if present, indicates that multiple
 * substitutions should be performed on each line, if applicable.
 * The usual escapes are supported as described in the regexp docs.
 */
int
dosub(lp, up, cmd, defpat)
LPTR	*lp, *up;
uchar	*cmd, *defpat;
{
	LINE	*cp;
	uchar	*pat, *sub;
	regexp	*prog;
	int	i, nsubs;
	bool_t	do_all;		/* do multiple substitutions per line */
	bool_t	do_div;		/* substituting string contains ^M */
	uchar	delim = *cmd;	/* delimiter (usual '/') */
	uchar	*q, *linestr;
	LINE	*newcp, *nextcp;
	bool_t	overflow = FALSE; /* line buffer overflow in substitute */
	static Char	subbuf[MAXLINE * 2];
#ifdef KANJI
	Char	*u, *usub;
#endif

	/*
	 * If no range was given, do the current line. If only one line
	 * was given, just do that one.
	 */
	if (lp->linep == NULL)
		*up = *lp = *Curschar;
	else {
		if (up->linep == NULL)
			*up = *lp;
	}

#ifndef NO_SAVE_HUGE
	if (defpat == NULL) {	/* Not invoked from g// command */
		u_save(lp->linep->prev, up->linep->next);
	}
#endif

	pat = ++cmd;		/* skip the initial '/' */

	while (*cmd) {
		if (*cmd == '\\')	/* next char is quoted */
			cmd += 2;
		else if (*cmd == delim)	/* delimiter */
		{
			*cmd++ = NUL;
			break;
#ifdef KANJI
		} else if (isKanji(*cmd)) {
			cmd++;
			cmd++;
#endif
		} else
			cmd++;		/* regular character */
	}

	if (*pat == NUL) {
		if (defpat)		/* Invoked from g// */
			pat = defpat;
		else {
			emsg(nullpat);
			return 0;
		}
	}

	sub = cmd;

	do_all = FALSE;
	do_div = FALSE;

	while (*cmd) {
		if (*cmd == '\\')	/* next char is quoted */
			cmd += 2;
		else if (*cmd == delim)	/* delimiter */
		{
			do_all = (cmd[1] == 'g');
			*cmd++ = NUL;
			break;
#ifdef KANJI
		} else if (isKanji(*cmd)) {
			cmd++;
			cmd++;
#endif
		} else if (*cmd == CR) {
			do_div = TRUE;
			*cmd = NL;
			cmd++;
		} else {
			cmd++;		/* regular character */
		}
	}

	reg_ic = P(P_IC);	/* set "ignore case" flag appropriately */

	if (reg_ic)
		pat = foldcase(pat);

#ifdef KANJI
	if ((prog = regcomp(normalize(mapstring(pat)))) == NULL) {
		emsg(invalid);
		return 0;
	}

	usub = usave(normalize(sub));
	if (usub == NULL) {
		free((char *)prog);
		return 0;
	}

	nsubs = 0;

	for (cp = lp->linep; cp != NULL ;cp = cp->next) {

		linestr = cp->s;
		if (reg_ic)
			linestr = foldcase(linestr);
		u = normalize(linestr);
		if (regexec(prog, u, TRUE)) {	/* a match on this line */
			Char	*ns, *sns, *p, *np, *snp;

			newcp = cp;	/* remember last matched line */

			/*
			 * Get double space for a buffer
			 * to do the substitution into.
			 */
			sns = ns = subbuf;
			*sns = NUL;

			i = 0;
			p = u;
			snp = np = normalize(cp->s);

			do {
				for (ns = sns; *ns ;ns++)
					;
				/*
				 * copy up to the part that matched
				 */
				while (p < prog->startp[0]) {
					*ns++ = *np++;
					p++;
				}

				if (ns - sns > MAXLINE - 1) {
					overflow = TRUE;
					goto abend;
				}
				regsub(prog, usub, ns);
				i++;

				/*
				 * continue searching after the match
				 */
				p = prog->endp[0];
				np = snp + (p - u);

			} while (regexec(prog, p, FALSE) &&
				 prog->startp[0] >= p && do_all);

			for (ns = sns; *ns ;ns++)
				;

			/*
			 * copy the rest of the line, that didn't match
			 */
			while (*p) {
				*ns++ = *np++;
				p++;
			}

			if (ns - sns > MAXLINE) {
				overflow = TRUE;
				goto abend;
			}
			*ns = NUL;

			q = strsave(unnormalize(sns));
			if (q == NULL) {
				nomem = TRUE;
				goto abend;
			}
			free(cp->s);		/* free the original line */
			cp->s = q;		/* and save the modified str */
			cp->size = strlen(cp->s) + 1;

			if (do_div) {
				nextcp = cp->next;
				if (!inslines(cp, FORWARD, cp->s, FALSE)) {
					nomem = TRUE;
					goto abend;
				}
				if (Curschar->linep != cp) {
					LPTR	savep;

					savep = *Curschar;
					Curschar->linep = cp;
					Curschar->index = 0;
					delline(1, FALSE);
					*Curschar = savep;
				} else
					delline(1, FALSE);
				newcp = cp = nextcp->prev;
			}

			nsubs += i;
			CHANGED;
		}
		if (cp == up->linep)
			break;
	}
	free((char *)usub);
#else
	if ((prog = regcomp(mapstring(pat))) == NULL) {
		emsg(invalid);
		return 0;
	}

	nsubs = 0;

	for (cp = lp->linep; cp != NULL ;cp = cp->next) {

		linestr = cp->s;
		if (reg_ic)
			linestr = foldcase(linestr);
		if (regexec(prog, linestr, TRUE)) {  /* a match on this line */
			uchar	*ns, *sns, *p, *np;

			newcp = cp;	/* remember last matched line */

			/*
			 * Get double space for a buffer
			 * to do the substitution into.
			 */
			sns = ns = subbuf;
			*sns = NUL;

			i = 0;
			p = linestr;
			np = cp->s;

			do {
				for (ns = sns; *ns ;ns++)
					;
				/*
				 * copy up to the part that matched
				 */
				while (p < prog->startp[0]) {
					*ns++ = *np++;
					p++;
				}

				if (ns - sns > MAXLINE - 1) {
					overflow = TRUE;
					goto abend;
				}
				regsub(prog, sub, ns);
				i++;

				/*
				 * continue searching after the match
				 */
				p = prog->endp[0];
				np = cp->s + (p - linestr);

			} while (regexec(prog, p, FALSE) &&
				 prog->startp[0] >= p && do_all);

			for (ns = sns; *ns ;ns++)
				;

			/*
			 * copy the rest of the line, that didn't match
			 */
			while (*p) {
				*ns++ = *np++;
				p++;
			}

			if (ns - sns > MAXLINE) {
				overflow = TRUE;
				goto abend;
			}
			*ns = NUL;

			ns = strsave(sns);
			if (ns == NULL) {
				nomem = TRUE;
				goto abend;
			}
			free(cp->s);		/* free the original line */
			cp->s = ns;		/* and save the modified str */
			cp->size = strlen(cp->s) + 1;

			if (do_div) {
				nextcp = cp->next;
				if (!inslines(cp, FORWARD, cp->s, FALSE)) {
					nomem = TRUE;
					goto abend;
				}
				if (Curschar->linep != cp) {
					LPTR	savep;

					savep = *Curschar;
					Curschar->linep = cp;
					Curschar->index = 0;
					delline(1, FALSE);
					*Curschar = savep;
				} else
					delline(1, FALSE);
				newcp = cp = nextcp->prev;
			}

			nsubs += i;
			CHANGED;
		}
		if (cp == up->linep)
			break;
	}
#endif

abend:
	if (nsubs) {
		Curschar->linep = newcp;
		Curschar->index = 0;
		beginline(TRUE);
		if (!defpat) {
			cursupdate();
			updatescreen();
		}
		if (overflow)
			emsg(ovrflw);
		else if (!nomem && nsubs >= P(P_RP))
			smsg("%d substitution%c", nsubs, (nsubs>1) ? 's' : ' ');
	} else if (defpat == NULL)	/* Not invoked from g// */
		emsg(overflow? ovrflw: nomatch);

	free((char *)prog);
	return nsubs;
}

/*
 * doglob(cmd)
 *
 * Execute a global command of the form:
 *
 * g/pattern/X
 *
 * where 'x' is a command character, currently one of the following:
 *
 * d	Delete all matching lines
 * p	Print all matching lines
 *
 * The command character (as well as the trailing slash) is optional, and
 * is assumed to be 'p' if missing.
 */
void
doglob(lp, up, cmd, invcond)
LPTR	*lp, *up;
uchar	*cmd;
int	invcond;
{
	LINE	*cp;
	uchar	*pat;
	regexp	*prog;
	int	ndone, nsubs = 0;
	uchar	cmdchar = NUL;	/* what to do with matching lines */
	uchar	delim = *cmd;	/* delimiter (usual '/') */
	uchar	*p, *linestr;
	bool_t	first = TRUE;
	uchar	*ncmd;		/* cont'd command */
	LPTR	cline;
	static uchar	ncmdbuf[MAXCLIN];

	/*
	 * If no range was given, do every line. If only one line
	 * was given, just do that one.
	 */
	if (lp->linep == NULL) {
		*lp = *Filemem;
		*up = *(prevline(Fileend));
	} else {
		if (up->linep == NULL)
			*up = *lp;
	}

	pat = ++cmd;		/* skip the initial '/' */

	while (*cmd) {
		if (*cmd == '\\')	/* next char is quoted */
			cmd += 2;
		else if (*cmd == delim)	/* delimiter */
		{
			cmdchar = cmd[1];
			*cmd++ = NUL;
			ncmd = cmd+1;
			break;
#ifdef KANJI
		} else if (isKanji(*cmd)) {
			cmd++;
			cmd++;
#endif
		} else
			cmd++;		/* regular character */
	}
	if (cmdchar == NUL)
		cmdchar = 'p';

	if (cmdchar != 'p')	/* Save text if cmd has side effects */
		u_save(lp->linep->prev, up->linep->next);

	reg_ic = P(P_IC);	/* set "ignore case" flag appropriately */

	if (reg_ic)
		pat = foldcase(pat);

	if (cmdchar != 'd' && cmdchar != 'p' && cmdchar != 's') {
		emsg(invcom);
		return;
	}

#ifdef KANJI
	if ((prog = regcomp(normalize(mapstring(pat)))) == NULL)
#else
	if ((prog = regcomp(mapstring(pat))) == NULL)
#endif
	{
		emsg(invalid);
		return;
	}

	ndone = 0;
	got_int = FALSE;

	for (cp = lp->linep; cp != NULL && !got_int ;cp = cp->next) {

		linestr = cp->s;
		if (reg_ic)
			linestr = foldcase(linestr);
#ifdef KANJI
		if ((regexec(prog, normalize(linestr), TRUE) != 0) ^ invcond)
#else
		if ((regexec(prog, linestr, TRUE) != 0) ^ invcond) /* a match on this line */
#endif
		{

			switch (cmdchar) {

			case 'd':		/* delete the line */
				if (Curschar->linep != cp) {
					LPTR	savep;

					savep = *Curschar;
					Curschar->linep = cp;
					Curschar->index = 0;
					delline(1, FALSE);
					*Curschar = savep;
				} else
					delline(1, FALSE);
				break;

			case 'p':		/* print the line */
				if (first) {
					gotocmd(FALSE, 0);
					outstr("\r\n");
					first = FALSE;
				}
				prt_line(cp->s);
				outstr("\r\n");
				break;
			case 's':	/* Substitute */
				cline.linep = cp;
				cline.index = 0;
				strcpy(ncmdbuf,ncmd);
				nsubs += dosub(&cline,&cline,ncmdbuf,pat);
				break;
			}
			ndone++;
		}
		if (cp == up->linep)
			break;
	}

	if (ndone) {
		switch (cmdchar) {

		case 'd':
			updatescreen();
			if (ndone >= P(P_RP) || got_int)
				smsg("%s%d fewer line%c",
					got_int ? "Interrupt: " : "",
					ndone,
					(ndone > 1) ? 's' : ' ');
			break;

		case 'p':
			wait_return();
			break;
		case 's':
			cursupdate();
			updatescreen();
			if (nsubs >= P(P_RP))
				smsg("%d substitution%c", nsubs, (nsubs>1) ? 's' : ' ');
			break;
		}
	} else {
		if (got_int)
			emsg(intrpt);
		else
			emsg(nomatch);
	}

	got_int = FALSE;
	free((char *)prog);
}

/*
 * Character Searches
 */

#ifdef KANJI
static unsigned int lastc = NUL; /* last character searched for */
#else
static uchar lastc = NUL;	/* last character searched for */
#endif
static int  lastcdir;		/* last direction of character search */
static int  lastctype;		/* last type of search ("find" or "to") */

/*
 * searchc(c, dir, type)
 *
 * Search for character 'c', in direction 'dir'. If type is 0, move to
 * the position of the character, otherwise move to just before the char.
 */
bool_t
searchc(c, dir, type)
#ifdef KANJI
unsigned int	c;
#else
uchar	c;
#endif
int	dir;
int	type;
{
	LPTR	save;

	save = *Curschar;	/* save position in case we fail */
	lastc = c;
	lastcdir = dir;
	lastctype = type;

	/*
	 * On 'to' searches, skip one to start with so we can repeat
	 * searches in the same direction and have it work right.
	 */
	if (type)
		(dir == FORWARD) ? oneright() : oneleft();

	while ( (dir == FORWARD) ? oneright() : oneleft() ) {
#ifdef KANJI
		if (kgchar(Curschar) == c)
#else
		if (gchar(Curschar) == c)
#endif
		{
			if (type)
				(dir == FORWARD) ? oneleft() : oneright();
			return TRUE;
		}
	}
	*Curschar = save;
	return FALSE;
}

bool_t
crepsearch(flag)
int	flag;
{
	int	dir = lastcdir;
	int	rval;

	if (lastc == NUL)
		return FALSE;

	rval = searchc(lastc, flag ? OTHERDIR(lastcdir) : lastcdir, lastctype);

	lastcdir = dir;		/* restore dir., since it may have changed */

	return rval;
}

/*
 * "Other" Searches
 */

/*
 * showmatch - move the cursor to the matching paren or brace
 */
LPTR *
showmatch()
{
	static	LPTR	pos;
	int	(*move)(), inc(), dec();
#ifdef KANJI
	int	kinc(), kdec();
#endif
	uchar	initc = gchar(Curschar);	/* initial char */
	uchar	findc;				/* terminating char */
	uchar	c;
	int	count = 0;

	pos = *Curschar;		/* set starting point */

	switch (initc) {

	case '(':
		findc = ')';
		move = inc;
		break;
	case ')':
		findc = '(';
		move = dec;
		break;
	case '{':
		findc = '}';
#ifdef KANJI
#ifdef SJIS
		move = kinc;
#else /*EUC or JIS*/
		move = inc;
#endif
#else
		move = inc;
#endif
		break;
	case '}':
		findc = '{';
#ifdef KANJI
#ifdef SJIS
		move = kdec;
#else /*EUC or JIS*/
		move = dec;
#endif
#else
		move = dec;
#endif
		break;
	case '[':
		findc = ']';
		move = inc;
		break;
	case ']':
		findc = '[';
		move = dec;
		break;
	default:
		return (LPTR *) NULL;
	}

	while ((*move)(&pos) != -1) {		/* until end of file */
		c = gchar(&pos);
		if (c == initc)
			count++;
		else if (c == findc) {
			if (count == 0)
				return &pos;
			count--;
		}
	}
	return (LPTR *) NULL;			/* never found it */
}

/*
 * findfunc(dir) - Find the next function in direction 'dir'
 *
 * Return TRUE if a function was found.
 */
bool_t
findfunc(dir)
int	dir;
{
	LPTR	*curr;

	curr = Curschar;

	do {
		curr = (dir == FORWARD) ? nextline(curr) : prevline(curr);

		if (curr != NULL && curr->linep->s[0] == '{') {
			setpcmark();
			*Curschar = *curr;
			return TRUE;
		}
	} while (curr != NULL);

	return FALSE;
}

/*
 * The following routines do the word searches performed by the
 * 'w', 'W', 'b', 'B', 'e', and 'E' commands.
 */

#ifdef KANJI

/*
 * To perform these searches, characters are placed into one of nine
 * classes, and transitions between classes determine word boundaries.
 *
 * The classes are:
 *
 * 0 - 1 or 2 byte white space
 * 1 - 1 byte katakana
 * 2 - 1 byte letters, digits, and underscore
 * 3 - 1 byte everything else
 * 4 - 2 byte kanji
 * 5 - 2 byte hiragana
 * 6 - 2 byte katakana (and cho-on)
 * 7 - 2 byte letters, digits
 * 8 - 2 byte everything else
 */

static	int	stype;		/* type of the word motion being performed */

/*
 * cls(c) - returns the class of character 'c'
 *
 * The 'type' of the current search modifies the classes of characters
 * if a 'W', 'B', or 'E' motion is being done. In this case, non-blank
 * characters are reported as class 1 since only white space boundaries
 * are of interest.
 */
static	int
cls(c)
unsigned int	c;
{
	if (c == ' ' || c == TAB || c == NUL ||
			c == (unsigned int)(KS1 << 8 | KS2))
		return 0;

	if (stype)
		return 1;

	if (c <= 0xff) {
		/*
		 * 1 byte characters
		 */
		if (c & 0x80)				/* katakana */
			return 1;
		else if (isalnum(c) || c == '_')	/* letters, digits */
			return 2;
		else
			return 3;
	}

	/*
	 * 2 byte characters
	 */

#ifdef SJIS
	if (c >= 0x889f)				/* kanji */
		return 4;
	else if (c >= 0x829f && c <= 0x82f1)		/* hiragana */
		return 5;
	else if (c >= 0x8340 && c <= 0x8396 || c == 0x815b) /* katakana */
		return 6;
	else if (c >= 0x824f && c <= 0x829a)		/* letters, digits */
		return 7;

#else /*EUC or JIS*/

	if (c >= 0xb0a1)				/* kanji */
		return 4;
	else if (c >= 0xa4a1 && c <= 0xa4f3)		/* hiragana */
		return 5;
	else if (c >= 0xa5a1 && c <= 0xa5f6 || c == 0xa1bc) /* katakana */
		return 6;
	else if (c >= 0xa3b0 && c <= 0xa3fa)		/* letters, digits */
		return 7;
#endif

	return 8;
}

static int
kinc2(p)
LPTR	*p;
{
	int	ret;

	if (gchar(p) == NUL)
		return kinc(p);
	if ((ret = kinc(p)) == 1) {
		if (kinc(p) == -1)
			ret = -1;
		kdec(p);
	}
	return ret;
}

/*
 * fwd_word(pos, type) - move forward one word
 *
 * Returns the resulting position, or NULL if EOF was reached.
 */
LPTR *
fwd_word(p, type)
LPTR	*p;
int	type;
{
	static	LPTR	pos;
	int	sclass;			/* starting class */

	pos = *p;

	stype = type;
	sclass = cls(kgchar(p));

	/*
	 * We always move at least one character.
	 */
	if (kinc2(&pos) == -1)
		return NULL;

	if (sclass != 0) {
		while (cls(kgchar(&pos)) == sclass) {
			if (kinc2(&pos) == -1)
				return &pos;
		}
		/*
		 * If we went from 1 -> 2 or 2 -> 1, return here.
		 */
		if (cls(kgchar(&pos)) != 0)
			return &pos;
	}

	/* We're in white space; go to next non-white */

	while (cls(kgchar(&pos)) == 0) {
		/*
		 * We'll stop if we land on a blank line
		 */
		if (pos.index == 0 && pos.linep->s[0] == NUL)
			break;

		if (kinc2(&pos) == -1)
			return &pos;
	}

	return &pos;
}

/*
 * bck_word(pos, type) - move backward one word
 *
 * Returns the resulting position, or NULL if EOF was reached.
 */
LPTR *
bck_word(p, type)
LPTR	*p;
int	type;
{
	static	LPTR	pos;
	int	sclass;			/* starting class */

	pos = *p;

	stype = type;
	sclass = cls(kgchar(p));

	if (kdec(&pos) == -1)
		return NULL;

	/*
	 * If we're in the middle of a word, we just have to
	 * back up to the start of it.
	 */
	if (cls(kgchar(&pos)) == sclass && sclass != 0) {
		/*
		 * Move backward to start of the current word
		 */
		while (cls(kgchar(&pos)) == sclass) {
			if (kdec(&pos) == -1)
				return &pos;
		}
		kinc(&pos);			/* overshot - forward one */
		return &pos;
	}

	/*
	 * We were at the start of a word. Go back to the start
	 * of the prior word.
	 */

	while (cls(kgchar(&pos)) == 0) {	/* skip any white space */
		/*
		 * We'll stop if we land on a blank line
		 */
		if (pos.index == 0 && pos.linep->s[0] == NUL)
			return &pos;

		if (kdec(&pos) == -1)
			return &pos;
	}

	sclass = cls(kgchar(&pos));

	/*
	 * Move backward to start of this word.
	 */
	while (cls(kgchar(&pos)) == sclass) {
		if (kdec(&pos) == -1)
			return &pos;
	}
	kinc(&pos);			/* overshot - forward one */

	return &pos;
}

/*
 * end_word(pos, type, in_change) - move to the end of the word
 *
 * There is an apparent bug in the 'e' motion of the real vi. At least
 * on the System V Release 3 version for the 80386. Unlike 'b' and 'w',
 * the 'e' motion crosses blank lines. When the real vi crosses a blank
 * line in an 'e' motion, the cursor is placed on the FIRST character
 * of the next non-blank line. The 'E' command, however, works correctly.
 * Since this appears to be a bug, I have not duplicated it here.
 *
 * There's a strange special case here that the 'in_change' parameter
 * helps us deal with. Vi effectively turns 'cw' into 'ce'. If we're on
 * a word with only one character, we need to stick at the current
 * position so we don't change two words.
 *
 * Returns the resulting position, or NULL if EOF was reached.
 */
LPTR *
end_word(p, type, in_change)
LPTR	*p;
int	type;
bool_t	in_change;
{
	static	LPTR	pos;
	int	sclass;			/* starting class */

	pos = *p;

	stype = type;
	sclass = cls(kgchar(p));

	if (kinc2(&pos) == -1)
		return NULL;

	/*
	 * If we're in the middle of a word, we just have to
	 * move to the end of it.
	 */
	if (cls(kgchar(&pos)) == sclass && sclass != 0) {
		/*
		 * Move forward to end of the current word
		 */
		while (cls(kgchar(&pos)) == sclass) {
			if (kinc2(&pos) == -1)
				return &pos;
		}
		kdec(&pos);			/* overshot - forward one */
		return &pos;
	}

	/*
	 * We were at the end of a word. Go to the end of the next
	 * word, unless we're doing a change. In that case we stick
	 * at the end of the current word.
	 */
	if (in_change)
		return p;

	while (cls(kgchar(&pos)) == 0) {	/* skip any white space */
		if (kinc2(&pos) == -1)
			return &pos;
	}

	sclass = cls(kgchar(&pos));

	/*
	 * Move forward to end of this word.
	 */
	while (cls(kgchar(&pos)) == sclass) {
		if (kinc2(&pos) == -1)
			return &pos;
	}
	kdec(&pos);			/* overshot - forward one */

	return &pos;
}

#else

/*
 * To perform these searches, characters are placed into one of three
 * classes, and transitions between classes determine word boundaries.
 *
 * The classes are:
 *
 * 0 - white space
 * 1 - letters, digits, and underscore
 * 2 - everything else
 */

static	int	stype;		/* type of the word motion being performed */

#define	C0(c)	(((c) == ' ') || ((c) == '\t') || ((c) == NUL))
#define	C1(c)	(isalpha(c) || isdigit(c) || ((c) == '_'))

/*
 * cls(c) - returns the class of character 'c'
 *
 * The 'type' of the current search modifies the classes of characters
 * if a 'W', 'B', or 'E' motion is being done. In this case, chars. from
 * class 2 are reported as class 1 since only white space boundaries are
 * of interest.
 */
static	int
cls(c)
uchar	c;
{
	if (C0(c))
		return 0;

	if (C1(c))
		return 1;

	/*
	 * If stype is non-zero, report these as class 1.
	 */
	return (stype == 0) ? 2 : 1;
}

static int
inc2(p)
LPTR	*p;
{
	int	ret;

	if (gchar(p) == NUL)
		return inc(p);
	if ((ret = inc(p)) == 1) {
		if (inc(p) == -1)
			ret = -1;
		dec(p);
	}
	return ret;
}

/*
 * fwd_word(pos, type) - move forward one word
 *
 * Returns the resulting position, or NULL if EOF was reached.
 */
LPTR *
fwd_word(p, type)
LPTR	*p;
int	type;
{
	static	LPTR	pos;
	int	sclass;			/* starting class */

	pos = *p;

	stype = type;
	sclass = cls(gchar(p));

	/*
	 * We always move at least one character.
	 */
	if (inc2(&pos) == -1)
		return NULL;

	if (sclass != 0) {
		while (cls(gchar(&pos)) == sclass) {
			if (inc2(&pos) == -1)
				return &pos;
		}
		/*
		 * If we went from 1 -> 2 or 2 -> 1, return here.
		 */
		if (cls(gchar(&pos)) != 0)
			return &pos;
	}

	/* We're in white space; go to next non-white */

	while (cls(gchar(&pos)) == 0) {
		/*
		 * We'll stop if we land on a blank line
		 */
		if (pos.index == 0 && pos.linep->s[0] == NUL)
			break;

		if (inc2(&pos) == -1)
			return &pos;
	}

	return &pos;
}

/*
 * bck_word(pos, type) - move backward one word
 *
 * Returns the resulting position, or NULL if EOF was reached.
 */
LPTR *
bck_word(p, type)
LPTR	*p;
int	type;
{
	static	LPTR	pos;
	int	sclass;			/* starting class */

	pos = *p;

	stype = type;
	sclass = cls(gchar(p));

	if (dec(&pos) == -1)
		return NULL;

	/*
	 * If we're in the middle of a word, we just have to
	 * back up to the start of it.
	 */
	if (cls(gchar(&pos)) == sclass && sclass != 0) {
		/*
		 * Move backward to start of the current word
		 */
		while (cls(gchar(&pos)) == sclass) {
			if (dec(&pos) == -1)
				return &pos;
		}
		inc(&pos);			/* overshot - forward one */
		return &pos;
	}

	/*
	 * We were at the start of a word. Go back to the start
	 * of the prior word.
	 */

	while (cls(gchar(&pos)) == 0) {		/* skip any white space */
		/*
		 * We'll stop if we land on a blank line
		 */
		if (pos.index == 0 && pos.linep->s[0] == NUL)
			return &pos;

		if (dec(&pos) == -1)
			return &pos;
	}

	sclass = cls(gchar(&pos));

	/*
	 * Move backward to start of this word.
	 */
	while (cls(gchar(&pos)) == sclass) {
		if (dec(&pos) == -1)
			return &pos;
	}
	inc(&pos);			/* overshot - forward one */

	return &pos;
}

/*
 * end_word(pos, type, in_change) - move to the end of the word
 *
 * There is an apparent bug in the 'e' motion of the real vi. At least
 * on the System V Release 3 version for the 80386. Unlike 'b' and 'w',
 * the 'e' motion crosses blank lines. When the real vi crosses a blank
 * line in an 'e' motion, the cursor is placed on the FIRST character
 * of the next non-blank line. The 'E' command, however, works correctly.
 * Since this appears to be a bug, I have not duplicated it here.
 *
 * There's a strange special case here that the 'in_change' parameter
 * helps us deal with. Vi effectively turns 'cw' into 'ce'. If we're on
 * a word with only one character, we need to stick at the current
 * position so we don't change two words.
 *
 * Returns the resulting position, or NULL if EOF was reached.
 */
LPTR *
end_word(p, type, in_change)
LPTR	*p;
int	type;
bool_t	in_change;
{
	static	LPTR	pos;
	int	sclass;			/* starting class */

	pos = *p;

	stype = type;
	sclass = cls(gchar(p));

	if (inc2(&pos) == -1)
		return NULL;

	/*
	 * If we're in the middle of a word, we just have to
	 * move to the end of it.
	 */
	if (cls(gchar(&pos)) == sclass && sclass != 0) {
		/*
		 * Move forward to end of the current word
		 */
		while (cls(gchar(&pos)) == sclass) {
			if (inc2(&pos) == -1)
				return &pos;
		}
		dec(&pos);			/* overshot - forward one */
		return &pos;
	}

	/*
	 * We were at the end of a word. Go to the end of the next
	 * word, unless we're doing a change. In that case we stick
	 * at the end of the current word.
	 */
	if (in_change)
		return p;

	while (cls(gchar(&pos)) == 0) {		/* skip any white space */
		if (inc2(&pos) == -1)
			return &pos;
	}

	sclass = cls(gchar(&pos));

	/*
	 * Move forward to end of this word.
	 */
	while (cls(gchar(&pos)) == sclass) {
		if (inc2(&pos) == -1)
			return &pos;
	}
	dec(&pos);			/* overshot - forward one */

	return &pos;
}

#endif
