/* $Header: /nw/tony/src/stevie/src/RCS/unix.c,v 1.8 89/08/06 09:51:13 tony Exp $
 *
 * System-dependent routines for UNIX System V or Berkeley.
 */

#include "stevie.h"
#ifdef BSD
#include <sgtty.h>
#include <sys/time.h>
#else
#include <termio.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>
#ifdef HAS_SYSDIR	/* BSD */
#include <sys/dir.h>
#endif
#ifdef HAS_DIRENT	/* most SYSV */
#include <dirent.h>
#endif

/*
 * maximum length of filename; usually defined in sys/dir.h or dirent.h
 */
#ifndef MAXNAMLEN
#define	MAXNAMLEN	14	/* SYSIII or earlier? */
#endif

#define	KEY_DELAY	300000	/* key sequence delay (in microseconds) */
#define DELAY_TIME	300000	/* shorttime delay (in microseconds) */

static	jmp_buf	onesec;

uchar	*Initfile = (uchar *)".stevierc";

#ifdef TERMCAP
extern	short	ospeed;
#endif

/*
 * character pushback stack
 */
static	int	npb = 0;
static	int	pbbuf[2];

/*
 * pushback() - subfunction to inchar()
 */
void
pushback(c)
	int	c;
{
	pbbuf[npb++] = c;
}

/*
 * inchar1() - subfunction to inchar()
 */
int
inchar1()
{
	uchar	c;

	if (npb > 0)
		return pbbuf[--npb];
	do {
		while (read(0, &c, 1) != 1)
			;
	} while (c == NUL);
	return c;
}

#ifdef KANJI

#ifdef JIS
/*
 * inchar() - get a character from the keyboard
 */
int
inchar()
{
	register int c, c1;
	void	(*savesig)(), onalarm();
	static	bool_t	kmode = FALSE;

	setoutkmode(FALSE);	/* exit kanji mode on output */
	flushbuf();		/* flush any pending output */
again:
	c = inchar1();
	if (c != ESC) {
		if (kmode  &&  c >= ' ')
			c = (c << 8) | inchar1() | 0x8080;
		got_int = FALSE;
		return c;
	}
	if (setjmp(onesec) != 0) {
		/*
		 * no more input
		 */
		signal(SIGALRM, savesig);
		got_int = FALSE;
		return ESC;
	}
	savesig = signal(SIGALRM, onalarm);
#ifdef BSD
	alarm_ms(KEY_DELAY);
#else
	alarm(1);
#endif
	c = inchar1();
	/*
	 * more input in one second
	 */
#ifdef BSD
	alarm_ms(0);
#else
	alarm(0);
#endif
	signal(SIGALRM, savesig);
	if (c == '$' || c == '(') {
		c1 = inchar1();
		if (c == '$' && (c1 == '@' || c1 == 'B')) {
			kmode = TRUE;
			goto again;
		}
		if (c == '(' && (c1 == 'B' || c1 == 'J' || c1 == 'H')) {
			kmode = FALSE;
			goto again;
		}
		pushback(c1);
	}
	if ((c == '[' || c == 'O') && State == NORMAL) {
		c1 = inchar1();
		got_int = FALSE;
		switch (c1) {
		case 'A': return 'k';
		case 'B': return 'j';
		case 'C': return 'l';
		case 'D': return 'h';
		default:  break;
		}
		pushback(c1);
	}
	pushback(c);
	got_int = FALSE;
	return ESC;
}

#else /*!JIS*/

/*
 * inchar() - get a character from the keyboard
 */
int
inchar()
{
	register int c, c1;
	void	(*savesig)(), onalarm();

	flushbuf();		/* flush any pending output */
again:
	c = inchar1();
#ifdef EUC
	if (c == SS2) {
		inchar1();
		beep();
		goto again;
	}
#endif
	if (c != ESC) {
		if (isKanji(c))
			c = (c << 8) | inchar1();
		got_int = FALSE;
		return c;
	}
	if (setjmp(onesec) != 0) {
		/*
		 * no more input
		 */
		signal(SIGALRM, savesig);
		got_int = FALSE;
		return ESC;
	}
	savesig = signal(SIGALRM, onalarm);
#ifdef BSD
	alarm_ms(KEY_DELAY);
#else
	alarm(1);
#endif
	c = inchar1();
	/*
	 * more input in one second
	 */
#ifdef BSD
	alarm_ms(0);
#else
	alarm(0);
#endif
	signal(SIGALRM, savesig);
	if ((c == '[' || c == 'O') && State == NORMAL) {
		c1 = inchar1();
		got_int = FALSE;
		switch (c1) {
		case 'A': return 'k';
		case 'B': return 'j';
		case 'C': return 'l';
		case 'D': return 'h';
		default:  break;
		}
		pushback(c1);
	}
	pushback(c);
	got_int = FALSE;
	return ESC;
}
#endif /*!JIS*/

static void
onalarm()
{
	signal(SIGALRM, SIG_IGN);
	longjmp(onesec, 1);
}

#else /*!KANJI*/

/*
 * inchar() - get a character from the keyboard
 */
int
inchar()
{
	int	c;

	flushbuf();		/* flush any pending output */
	c = inchar1();
	got_int = FALSE;
	return c;
}
#endif /*!KANJI*/

static	uchar	outbuf[OBSIZE];
static	int	bpos = 0;

void
flushbuf()
{
	uchar * p = outbuf;
	int writed;

	while (bpos) {
		writed = write(2, p, bpos);
		if (writed < 0)
			continue;
		bpos -= writed;
		p += writed;
	}
}

/*
 * Macro to output a character. Used within this file for speed.
 */
#define	outone(c)	outbuf[bpos++] = c; if (bpos >= OBSIZE) flushbuf()

#ifndef JIS

/*
 * Function version for use outside this file.
 */
void
outchar(c)
uchar	c;
{
	outone(c);
}

void
outstr(s)
register uchar	*s;
{
	while (*s) {
		outone(*s++);
	}
}

void
beep()
{
	outone('\007');
}

#else /*JIS*/

static	bool_t	outkmode = FALSE;

/*
 * Function to control output kanji mode
 */
int
setoutkmode(newmode)
bool_t	newmode;
{
	if (outkmode == TRUE && newmode == FALSE) {
		outone(ESC);
		outone('(');
		outone(RFC);
		outkmode = FALSE;
	} else if (outkmode == FALSE && newmode == TRUE) {
		outone(ESC);
		outone('$');
		outone(KFC);
		outkmode = TRUE;
	}
	return outkmode;
}

/*
 * Function to output character
 */
void
outchar(c)
uchar	c;
{
	static	int	csave = NUL;

	if (outkmode) {
		if (csave) {
			outone(csave & 0x7f);
			outone(c & 0x7f);
			csave = NUL;
		} else if (isKanji(c)) {
			csave = c;
		} else {
			setoutkmode(FALSE);
			outone(c);
		}
	} else {
		if (isKanji(c)) {
			setoutkmode(TRUE);
			csave = c;
		} else {
			outone(c);
		}
	}
}

void
outstr(s)
register uchar	*s;
{
	while (*s) {
		outchar(*s++);
	}
}

void
beep()
{
	outchar('\007');
}

#endif /*JIS*/

void
delay()
{
	flushbuf();		/* flush any pending output */
#ifdef BSD
	sleep_ms(DELAY_TIME);
#else
	sleep(1);		/* may be longer */
#endif
}

#ifdef BSD
static	struct	sgttyb	ostate;
static	struct	ltchars oltchars;
#else
static	struct	termio	ostate;
#endif

/*
 * Go into cbreak mode
 */
void
set_tty()
{
#ifdef BSD
	struct	sgttyb	nstate;
	struct	ltchars	nltchars;

	ioctl(0, TIOCGETP, &ostate);
	ioctl(0, TIOCGLTC, &oltchars);
	nstate   = ostate;
	nltchars = oltchars;
	nstate.sg_flags  &= ~(XTABS|CRMOD|ECHO);
	nstate.sg_flags  |= CBREAK;
	nltchars.t_dsuspc = 0xff;
	nltchars.t_rprntc = 0xff;
	nltchars.t_flushc = 0xff;
	nltchars.t_werasc = 0xff;
	nltchars.t_lnextc = 0xff;
	ioctl(0, TIOCSETN, &nstate);
	ioctl(0, TIOCSLTC, &nltchars);
#ifdef TERMCAP
	ospeed = ostate.sg_ospeed;
#endif
#else
	struct	termio	nstate;

	ioctl(0, TCGETA, &ostate);
	nstate = ostate;
	nstate.c_lflag &= ~(ICANON|ECHO|ECHOE|ECHOK|ECHONL);
	nstate.c_cc[VMIN]   = 1;
	nstate.c_cc[VTIME]  = 0;
	ioctl(0, TCSETAW, &nstate);
#ifdef TERMCAP
	ospeed = ostate.c_cflag & CBAUD;
#endif
#endif
}

/*
 * Restore original terminal modes
 */
void
reset_tty()
{
#ifdef BSD
	ioctl(0, TIOCSETP, &ostate);
	ioctl(0, TIOCSLTC, &oltchars);
#else
	ioctl(0, TCSETAW, &ostate);
#endif
}

void
sig()
{
	signal(SIGINT, sig);
	signal(SIGQUIT, sig);

	got_int = TRUE;
}

void
onhup()
{
	FILE	*fopen(), *fopenb();
	register FILE	*fp;
	register LPTR	*lp;
	char	*Savefile, *getenv(), *mktemp();
#ifdef JIS
	register uchar	*p;
	bool_t	kmode = FALSE;
#endif

	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);
	signal(SIGHUP, SIG_IGN);
	signal(SIGTERM, SIG_IGN);
#ifdef SIGTSTP
	signal(SIGTSTP, SIG_IGN);
#endif
	if (!Changed)
		windexit(1);

	Savefile = strcat(mktemp(mkpath(getenv("HOME"), "viXXXXXX")), ".HUP");
	fp = fopenb(Savefile, "w");
	if (fp != NULL) {
		for (lp = Filemem; lp != NULL; lp = nextline(lp)) {
#ifdef JIS
			p = lp->linep->s;
			for (;;) {
				if (kmode && !(*p & 0x80)) {
					putc(ESC, fp);
					putc('(', fp);
					putc(RFC, fp);
					kmode = FALSE;
				} else if (!kmode && (*p & 0x80)) {
					putc(ESC, fp);
					putc('$', fp);
					putc(KFC, fp);
					kmode = TRUE;
				}
				if (*p) {
					putc((*p++ & 0x7f), fp);
					continue;
				}
				if (P(P_CR))
					putc(CR, fp);
				putc(NL, fp);
				break;
			}
#else
			fprintf(fp, "%s", lp->linep->s);
			fprintf(fp, P(P_CR)? "\r\n": "\n");
#endif
		}
		fclose(fp);
	}
	windexit(1);
}

#ifdef SIGTSTP
void
tstp()
{
	signal(SIGTSTP, SIG_IGN);
	windgoto(Rows-1, 0);
	if (T_TE) toutstr(T_TE, 1);
	flushbuf();
	reset_tty();
	signal(SIGTSTP, SIG_DFL);
	if (P(P_AW)) {
		if (!autowriteit(FALSE))
			goto tstp1;
	}
#ifdef BSD
#define	MASK(a)	(1 << ((a) - 1))
	sigsetmask(sigblock(0) & ~MASK(SIGTSTP));
#endif
	kill(getpid(), SIGTSTP);
	/*
	 * zzz...
	 */
tstp1:
	signal(SIGTSTP, tstp);
	set_tty();
	if (got_int)
		windexit(0);
	if (T_TI) toutstr(T_TI, 1);
	screenclear();
	updatescreen();
	windgoto(Cursrow, Curscol);
	flushbuf();
}
#endif

void
tstpon()
{
#ifdef SIGTSTP
	signal(SIGTSTP, tstp);
#else
	/* nothing to do here */
#endif
}

void
tstpoff()
{
#ifdef SIGTSTP
	signal(SIGTSTP, SIG_IGN);
#else
	/* nothing to do here */
#endif
}

void
windinit()
{
#ifdef	TERMCAP
	switch (t_init()) {
	case 1:
		break;
	case 0:
		fprintf(stderr, "insufficient terminal capability\n");
		exit(1);
	case -1:
		fprintf(stderr, "unknown terminal\n");
		exit(1);
	}
#else
	Columns = 80;
	P(P_LI) = Rows = 24;
#endif

	/*
	 * The code here makes sure that there isn't a window during which
	 * we could get interrupted and exit with the tty in a weird state.
	 */
	signal(SIGINT, sig);
	signal(SIGQUIT, sig);
	signal(SIGHUP, onhup);
	signal(SIGTERM, onhup);
#ifdef SIGTSTP
	signal(SIGTSTP, tstp);
#endif

	set_tty();

	if (got_int)
		windexit(0);

	if (T_TI) toutstr(T_TI, 1);
}

void
windexit(r)
int r;
{
	if (T_TE) toutstr(T_TE, 1);
	flushbuf();
	reset_tty();
	exit(r);
}

void
windgoto(r, c)
register int	r, c;
{
#ifdef	TERMCAP
	char	*tgoto();

	toutstr(tgoto(T_CM, c, r), 1);
#else
	r++;
	c++;

	/*
	 * Check for overflow once, to save time.
	 */
	if (bpos + 8 >= OBSIZE)
		flushbuf();

	outbuf[bpos++] = '\033';
	outbuf[bpos++] = '[';
	if (r >= 10) {
		outbuf[bpos++] = r / 10 + '0';
		outbuf[bpos++] = r % 10 + '0';
	} else
		outbuf[bpos++] = r + '0';
	outbuf[bpos++] = ';';
	if (c >= 10) {
		outbuf[bpos++] = c / 10 + '0';
		outbuf[bpos++] = c % 10 + '0';
	} else
		outbuf[bpos++] = c + '0';
	outbuf[bpos++] = 'H';
#endif
}

FILE *
fopenb(fname, mode)
uchar	*fname;
uchar	*mode;
{
	return fopen(fname, mode);
}

int
filestat(fname)
uchar	*fname;
{
	int	fd, ret;
	struct stat stbuf;
	extern int errno;

	fd = open(fname, 0);
	if (fd == -1) {
		if (errno == ENOENT)
			ret = 0;
		else
			ret = FS_EXIST;
	} else {
		ret = FS_EXIST | FS_READ;
		fstat(fd, &stbuf);
		if (stbuf.st_mode & S_IFDIR)
			ret |= FS_DIR;
		close(fd);
		if (access(fname, 2) == 0)
			ret |= FS_WRITE;
	}
	return ret;
}

/*
 * remove(file) - remove a file
 */
int
remove(file)
uchar *file;
{
	return unlink(file);
}

/*
 * Rename(of, nf) - rename existing file 'of' to 'nf'
 */
int
Rename(of, nf)
uchar	*of, *nf;
{
	unlink(nf);
	if (link(of, nf) != -1 && unlink(of) != -1)
		return 0;
	return -1;
}

/*
 * backupname(fname, ext) - generate filename for backup
 */
uchar *
backupname(fname, ext)
uchar	*fname, *ext;
{
	register uchar	*s, *top;
	int		elen;
	static uchar	bname[MAXPLEN];

	elen = strlen(ext);
	strcpy(bname, fname);
	for (s = bname; *s; s++)
		/* void */;
	while (s > bname && *s != '/')
		s--;
	top = ++s;
	while (*s && (s - top) + elen < MAXNAMLEN)
		s++;
	*s = NUL;
	strcat(bname, ext);
	return bname;
}

uchar *
fixname(s)
uchar	*s;
{
	return s;
}

uchar *
mkpath(dir, file)
uchar	*dir, *file;
{
	int	n;
	static uchar	buf[MAXPLEN];

	n = 0;
	if (dir) {
		strcpy(buf, dir);
		n = strlen(buf);
		if (buf[n-1] != '/')
			buf[n++] = '/';
	}
	strcpy(buf+n, file);
	return buf;
}

/*
 * doshell() - run a command or an interactive shell
 */
void
doshell(cmd)
uchar	*cmd;
{
	char	*getenv();
	uchar	cline[MAXCLIN];

	outstr("\r\n");
	flushbuf();

	if (cmd == NULL) {
		if ((cmd = (uchar *)getenv("SHELL")) == NULL)
			cmd = (uchar *)"/bin/sh";
		sprintf(cline, "%s -i", cmd);
		cmd = cline;
	}

	reset_tty();
	system(cmd);
	set_tty();

	wait_return();
}

void
cmdread(cmd)
uchar	*cmd;
{
	int fd, old, err;
	char tmpfile[MAXPLEN];

	if (cmd[0] == NUL) {
		emsg("Exec failed");
		return;
	}
	tmpnam(tmpfile);
	if ((fd = creat(tmpfile, 0600)) < 0) {
		emsg("Can't create temporary file");
		return;
	}
	old = dup(1);
#ifdef BSD
	dup2(fd, 1);
#else
	close(1);
	dup(fd);
#endif
	close(fd);
	outstr("\r");
	flushbuf();
	reset_tty();
	err = system(cmd);
	set_tty();
	close(1);
#ifdef BSD
	dup2(old, 1);
#else
	close(1);
	dup(old);
#endif
	if (err < 0)
		emsg("Exec failed");
	else {
		if (readfile(tmpfile, Curschar, 1))
			emsg("Can't open temporary file");
		else {
			updatescreen();
			CHANGED;
		}
	}
	remove(tmpfile);
}

#ifdef BSD
/*
 * sleep_ms: delay timer  by A.ITO
 * slightly modified by Junn Ohta
 */

/*
 * alarm in microsecond
 */
static
alarm_ms(n)
int	n;
{
	struct itimerval itval;

	itval.it_interval.tv_sec = 0;
	itval.it_interval.tv_usec = 0;
	itval.it_value.tv_sec = n / 1000000;
	itval.it_value.tv_usec = n % 1000000;
	setitimer(ITIMER_REAL, &itval, NULL);
}

/*
 * sleep in microsecond
 */
static
sleep_ms(n)
int	n;
{
	void	break_sleep();

	signal(SIGALRM, break_sleep);
	alarm_ms(n);
	pause();
	signal(SIGALRM, SIG_DFL);
}

static void
break_sleep()
{
}
#endif
