#
#	Makefile for UNIX
#

#######################################################
# For BSD
SYS	= UNIX_BSD
DIR	= HAS_SYSDIR
LIBS	= -ltermcap

#######################################################
# For System V
#SYS	= UNIX_SYSV
#DIR	= HAS_DIRENT
#LIBS	= -lcurses

#######################################################
# Character handling (select one from below)
#CHAR	= SJIS
#CHAR	= EUC
CHAR	= JIS
#CHAR	= BIT8
#CHAR	= NOKANJI

CFLAGS	= -O -D$(SYS) -D$(DIR) -D$(CHAR)
LDFLAGS	=

HDRS =	ascii.h env.h keymap.h ops.h param.h \
	regexp.h regmagic.h stevie.h term.h

OBJS =	alloc.o cmdline.o edit.o fileio.o help.o hexchars.o \
	linefunc.o main.o mark.o misccmds.o normal.o ops.o \
	param.o ptrfunc.o regexp.o regsub.o screen.o search.o \
	term.o undo.o version.o glob.o unix.o

SRCS =	$(OBJS:.o=.c)

all: stevie

stevie: $(OBJS)
	$(CC) $(LDFLAGS) -o stevie $(OBJS) $(LIBS)

lint:
	lint $(SRCS)

tags:
	ctags $(SRCS) $(HDRS)

stevie.doc: stevie.mm
	nroff -rB1 -Tlp -mm stevie.mm > stevie.doc

print:
	@pr $(HDRS) $(SRCS)

cflow:
	cflow $(SRCS) > cflow.for
	cflow -r $(SRCS) > cflow.rev

clean:
	rm -f $(OBJS)

main.o:		stevie.h env.h ascii.h param.h
alloc.o:	stevie.h env.h ascii.h param.h
cmdline.o:	stevie.h env.h ascii.h param.h term.h
edit.o:		stevie.h env.h ascii.h param.h term.h
fileio.o:	stevie.h env.h ascii.h param.h
help.o:		stevie.h env.h ascii.h param.h term.h
hexchars.o:	stevie.h env.h ascii.h param.h
linefunc.o:	stevie.h env.h ascii.h param.h
mark.o:		stevie.h env.h ascii.h param.h
misccmds.o:	stevie.h env.h ascii.h param.h term.h
normal.o:	stevie.h env.h ascii.h keymap.h param.h ops.h
ops.o:		stevie.h env.h ascii.h param.h ops.h
param.o:	stevie.h env.h ascii.h param.h
ptrfunc.o:	stevie.h env.h ascii.h param.h
regexp.o:	env.h regexp.h regmagic.h
regsub.o:	regexp.h regmagic.h
screen.o:	stevie.h env.h ascii.h param.h term.h
search.o:	stevie.h env.h ascii.h param.h regexp.h
term.o:		stevie.h env.h ascii.h param.h term.h
undo.o:		stevie.h env.h ascii.h param.h
unix.o:		stevie.h env.h ascii.h param.h term.h
glob.o:		env.h
